@@ -42,7 +42,7 @@
 ;;    - :fg and :bg take a color as an argument, which can either be a numeric
 ;;      index into the color map or a hexadecimal color in the form of "#fff"
 ;;      or "#ffffff".
-;;    - :reverse and :bright take either t or nil as an argument. T enables
+;;    - :reverse and :bright take either t or nil as an argument. A t enables
 ;;      the setting and nil disables it.
 ;;    - :push and :pop take no arguments. :push pushes the current settings onto
 ;;      the color stack, leaving the current settings intact. :pop pops color
@@ -77,7 +77,8 @@ then call (update-color-map).")
   (labels ((max-min (x y) (max 0 (min 1 (+ x y)))))
     (setf (xlib:color-red color) (max-min (xlib:color-red color) amt)
           (xlib:color-green color) (max-min (xlib:color-green color) amt)
-          (xlib:color-blue color) (max-min (xlib:color-blue color) amt))))
+          (xlib:color-blue color) (max-min (xlib:color-blue color) amt))
+    color))
 
 (defun hex-to-xlib-color (color)
   (cond
@@ -106,21 +107,43 @@ then call (update-color-map).")
                           color))))
 
 (defun alloc-color (screen color)
-  (let ((colormap (xlib:screen-default-colormap (screen-number screen))))
-    (cond ((typep color 'xlib:color) (xlib:alloc-color colormap color))
-          (t (xlib:alloc-color colormap (lookup-color screen color))))))
+  ;; We add an alpha channel to the color returned by xlib:alloc-color
+  (logior (xlib:alloc-color (xlib:screen-default-colormap (screen-number screen))
+                            (lookup-color screen color))
+          (ash #xff 24)))
 
 ;; Normal colors are dimmed and bright colors are intensified in order
 ;; to more closely resemble the VGA pallet.
 (defun update-color-map (screen)
   "Read *colors* and cache their pixel colors for use when rendering colored text."
-  (labels ((map-colors (amt)
+  (labels ((map-colors (adj)
              (loop for c in *colors*
-                   as color = (lookup-color screen c)
-                   do (adjust-color color amt)
+                   as color = (typecase c
+                                ;; If the color element is a list, use
+                                ;; the first or second color in the
+                                ;; list appropriately
+                                (cons
+                                 (lookup-color
+                                  screen
+                                  (case adj
+                                    (:normal
+                                     (first c))
+                                    (:bright
+                                     (second c)))))
+                                ;; If the color element is not a list,
+                                ;; look up the color and adjust it
+                                ;; automatically
+                                (t
+                                 (adjust-color
+                                  (lookup-color screen c)
+                                  (case adj
+                                    (:normal
+                                     -0.25)
+                                    (:bright
+                                     0.25)))))
                    collect (alloc-color screen color))))
-    (setf (screen-color-map-normal screen) (apply #'vector (map-colors -0.25))
-          (screen-color-map-bright screen) (apply #'vector (map-colors 0.25)))))
+    (setf (screen-color-map-normal screen) (apply #'vector (map-colors :normal))
+          (screen-color-map-bright screen) (apply #'vector (map-colors :bright)))))
 
 (defun update-screen-color-context (screen)
   (let* ((cc (screen-message-cc screen))
@@ -195,7 +218,7 @@ If COLOR isn't a colorcode a list containing COLOR is returned."
                     (screen-color-map-bright screen)
                     (screen-color-map-normal screen))
                 color))
-        (t (alloc-color screen color))))
+        (t (nth-value 0 (alloc-color screen color)))))
 
 (defun find-font (cc specified-font &aux (font (or specified-font 0)))
   (if (integerp font)
@@ -204,6 +227,11 @@ If COLOR isn't a colorcode a list containing COLOR is returned."
 
 (defgeneric apply-color (ccontext modifier &rest arguments))
 
+(defmethod apply-color :around ((cc ccontext) modifier &rest arguments)
+  (declare (ignorable cc modifier arguments))
+  (when *draw-in-color*
+    (call-next-method)))
+
 (defmethod apply-color ((cc ccontext) (modifier (eql :fg)) &rest args)
   (setf (ccontext-fg cc) (first args))
   (let* ((gcontext (ccontext-gc cc))
@@ -269,6 +297,13 @@ If COLOR isn't a colorcode a list containing COLOR is returned."
   ;; rendering. Since it doesn't otherwise have any effects, we just ignore it.
   (declare (ignore cc modifier args)))
 
+;; Two more special cases. 
+(defmethod apply-color ((cc ccontext) (modifier (eql :on-click)) &rest args)
+  (declare (ignore cc modifier args)))
+
+(defmethod apply-color ((cc ccontext) (modifier (eql :on-click-end)) &rest args)
+  (declare (ignore cc modifier args)))
+
 (defun max-font-height (parts cc)
   "Return the biggest font height for all of the fonts occurring in PARTS in
 the form of (:FONT ...) modifiers."
@@ -316,43 +351,69 @@ string."
                   (reset-color-context cc)
                   (return (values max-width height)))))
 
-(defun render-string (string-or-parts cc x y &aux (draw-x x))
+(defun render-string (string-or-parts cc x y &key ml &aux (draw-x x))
   "Renders STRING-OR-PARTS to the pixmap in CC. Returns the height and width of
 the rendered line as two values. The returned width is the value of X plus the
 rendered width."
-  (let* ((parts (if (stringp string-or-parts)
-                    (parse-color-string string-or-parts)
-                    string-or-parts))
-         (height (max-font-height parts cc)))
-    (loop
-       for (part . rest) on parts
-       for font-height-difference = (- height
-                                       (font-height (ccontext-font cc)))
-       for y-to-center = (floor (/ font-height-difference 2))
-       if (stringp part)
-       do (draw-image-glyphs
-           (ccontext-px cc)
-           (ccontext-gc cc)
-           (ccontext-font cc)
-           draw-x (+ y y-to-center (font-ascent (ccontext-font cc)))
-           part
-           :translate #'translate-id
-           :size 16)
-         (incf draw-x (text-line-width (ccontext-font cc)
-                                       part
-                                       :translate #'translate-id))
-       else
-       do (if (eq :> (first part))
-              (progn (render-string rest cc
-                                    (- (xlib:drawable-width (ccontext-px cc))
-                                       x
-                                       (rendered-string-size rest cc))
-                                    y)
-                     (loop-finish))
-              (apply #'apply-color cc (first part) (rest part))))
-    (values height draw-x)))
+  (macrolet ((register (thing)
+               `(let ((top ,thing))
+                  (when top
+                    (register-ml-boundaries-with-id ml
+                                                    (first top)
+                                                    draw-x
+                                                    y
+                                                    (+ y y-to-center
+                                                       (font-ascent
+                                                        (ccontext-font cc)))
+                                                    (second top)
+                                                    (third top))))))
+    (let* ((parts (if (stringp string-or-parts)
+                      (parse-color-string string-or-parts)
+                      string-or-parts))
+           (height (max-font-height parts cc))
+           (current-on-click nil))
+      (loop
+        for (part . rest) on parts
+        for font-height-difference = (- height
+                                        (font-height (ccontext-font cc)))
+        for y-to-center = (floor (/ font-height-difference 2))
+        if (stringp part)
+          do (draw-image-glyphs
+              (ccontext-px cc)
+              (ccontext-gc cc)
+              (ccontext-font cc)
+              draw-x (+ y y-to-center (font-ascent (ccontext-font cc)))
+              part
+              :translate #'translate-id
+              :size 16)
+             (incf draw-x (text-line-width (ccontext-font cc)
+                                           part
+                                           :translate #'translate-id))
+        else
+          do (case (first part)
+               ((:on-click)
+                (when ml
+                  (push (list draw-x (cadr part) (cddr part)) current-on-click)))
+               ((:on-click-end)
+                (when ml
+                  (register (pop current-on-click))))
+               ((:>)
+                (let ((xbeg (- (xlib:drawable-width (ccontext-px cc))
+                               x
+                               (rendered-string-size rest cc))))
+                  ;; Terminate all clickable areas as they cannot cross the :>
+                  ;; boundary.
+                  (when ml
+                    (loop for top = (pop current-on-click)
+                          while top
+                          do (register top)))
+                  (render-string rest cc xbeg y :ml ml))
+                (loop-finish))
+               (otherwise
+                (apply #'apply-color cc (first part) (rest part)))))
+      (values height draw-x))))
 
-(defun render-strings (cc padx pady strings highlights)
+(defun render-strings (cc padx pady strings highlights &key ml)
   (let* ((gc (ccontext-gc cc))
          (xwin (ccontext-win cc))
          (px (ccontext-px cc))
@@ -382,17 +443,17 @@ rendered width."
        for row from 0 to (length strings)
        for line-height = (max-font-height parts cc)
        if (find row highlights :test 'eql)
-       do (xlib:draw-rectangle px gc 0 y (xlib:drawable-width px) line-height t)
+       do (xlib:draw-rectangle px gc 0 (+ pady y) (xlib:drawable-width px) line-height t)
          (xlib:with-gcontext (gc :foreground (xlib:gcontext-background gc)
                                  :background (xlib:gcontext-foreground gc))
            ;; If we don't switch the default colors, a color operation
            ;; resetting either color to its default value would undo the
            ;; switch.
            (rotatef (ccontext-default-fg cc) (ccontext-default-bg cc))
-           (render-string parts cc (+ padx 0) (+ pady y))
+           (render-string parts cc (+ padx 0) (+ pady y) :ml ml)
            (rotatef (ccontext-default-fg cc) (ccontext-default-bg cc)))
        else
-       do (render-string parts cc (+ padx 0) (+ pady y))
+         do (render-string parts cc (+ padx 0) (+ pady y) :ml ml)
        end
        do (incf y line-height))
     (xlib:copy-area px gc 0 0
