@@ -24,6 +24,8 @@
 
 (in-package #:stumpwm)
 
+(export '(*button-state*))
+
 ;;; Event handler functions
 
 (defparameter *event-fn-table* (make-hash-table)
@@ -32,21 +34,25 @@
 (defvar *current-event-time* nil)
 
 (defmacro define-stump-event-handler (event keys &body body)
-  (let ((fn-name (gensym))
-        (event-slots (gensym)))
-    `(labels ((,fn-name (&rest ,event-slots &key ,@keys &allow-other-keys)
-               (declare (ignore ,event-slots))
-               ,@body))
-      (setf (gethash ,event *event-fn-table*) #',fn-name))))
+  (let ((event-slots (gensym)))
+    (multiple-value-bind (body declarations docstring)
+        (parse-body body :documentation t)
+      `(setf (gethash ,event *event-fn-table*)
+             (lambda (&rest ,event-slots &key ,@keys &allow-other-keys)
+               (declare (ignore ,event-slots)
+                        ,@(cdar declarations))
+               ,@(when docstring
+                   (list docstring))
+               ,@body)))))
 
 ;;; Configure request
 
-(flet ((has-x (mask) (= 1 (logand mask 1)))
-       (has-y (mask) (= 2 (logand mask 2)))
-       (has-w (mask) (= 4 (logand mask 4)))
-       (has-h (mask) (= 8 (logand mask 8)))
-       (has-bw (mask) (= 16 (logand mask 16)))
-       (has-stackmode (mask) (= 64 (logand mask 64))))
+(flet ((has-x (mask) (logbitp 0 mask))
+       (has-y (mask) (logbitp 1 mask))
+       (has-w (mask) (logbitp 2 mask))
+       (has-h (mask) (logbitp 3 mask))
+       (has-bw (mask) (logbitp 4 mask))
+       (has-stackmode (mask) (logbitp 6 mask)))
   (defun configure-managed-window (win x y width height stack-mode value-mask)
     ;; Grant the configure request but then maximize the window after the
     ;; granting.
@@ -59,7 +65,7 @@
       (when (window-urgent-p win)
         (window-clear-urgency win)))
     (when (or (has-x value-mask) (has-y value-mask))
-      (group-move-request (window-group win) win x y :parent))
+      (group-move-request (window-group win) win x y :root))
     (when (or (has-w value-mask) (has-h value-mask))
       (group-resize-request (window-group win) win width height))
     (when (has-stackmode value-mask)
@@ -86,25 +92,24 @@
 
 (define-stump-event-handler :configure-request (stack-mode #|parent|# window #|above-sibling|# x y width height border-width value-mask)
   (dformat 3 "CONFIGURE REQUEST ~@{~S ~}~%" stack-mode window x y width height border-width value-mask)
-  (let ((win (find-window window)))
-    (if win
-        (configure-managed-window win x y width height stack-mode value-mask)
-        (configure-unmanaged-window window x y width height border-width value-mask))))
+  (if-let ((win (find-window window)))
+    (configure-managed-window win x y width height stack-mode value-mask)
+    (configure-unmanaged-window window x y width height border-width value-mask)))
 
 (define-stump-event-handler :configure-notify (stack-mode #|parent|# window #|above-sibling|# x y width height border-width value-mask)
   (dformat 4 "CONFIGURE NOTIFY ~@{~S ~}~%" stack-mode window x y width height border-width value-mask)
-  (let ((screen (find-screen window)))
-    (when screen
-      (let ((old-heads (screen-heads screen))
-            (new-heads (make-screen-heads screen (screen-root screen))))
-        (cond
-          ((equalp old-heads new-heads)
-           (dformat 3 "Bogus configure-notify on root window of ~S~%" screen) t)
-          (t
-           (dformat 1 "Updating Xinerama configuration for ~S.~%" screen)
-           (if new-heads
-               (head-force-refresh screen new-heads)
-               (dformat 1 "Invalid configuration! ~S~%" new-heads))))))))
+  (when-let ((screen (find-screen window)))
+    (let ((old-heads (screen-heads screen))
+          (new-heads (make-screen-heads screen (screen-root screen))))
+      (cond
+        ((equalp old-heads new-heads)
+         (dformat 3 "Bogus configure-notify on root window of ~S~%" screen) t)
+        (t
+         (dformat 1 "Updating Xrandr or Xinerama configuration for ~S.~%" screen)
+         (if new-heads
+             (progn (head-force-refresh screen new-heads)
+                    (update-mode-lines screen))
+             (dformat 1 "Invalid configuration! ~S~%" new-heads)))))))
 
 (define-stump-event-handler :map-request (parent send-event-p window)
   (unless send-event-p
@@ -145,15 +150,14 @@
   (dformat 2 "UNMAP: ~s ~s ~a~%" send-event-p (not (xlib:window-equal event-window window)) (find-window window))
   (unless (and (not send-event-p)
                (not (xlib:window-equal event-window window)))
-    (let ((window (find-window window)))
-      ;; if we can't find the window then there's nothing we need to
-      ;; do.
-      (when window
-        (if (plusp (window-unmap-ignores window))
-            (progn
-              (dformat 3 "decrement ignores! ~d~%" (window-unmap-ignores window))
-              (decf (window-unmap-ignores window)))
-            (withdraw-window window))))))
+    ;; if we can't find the window then there's nothing we need to
+    ;; do.
+    (when-let ((window (find-window window)))
+      (if (plusp (window-unmap-ignores window))
+          (progn
+            (dformat 3 "decrement ignores! ~d~%" (window-unmap-ignores window))
+            (decf (window-unmap-ignores window)))
+          (withdraw-window window)))))
 
 (define-stump-event-handler :destroy-notify (send-event-p event-window window)
   (unless (or send-event-p
@@ -161,13 +165,11 @@
     ;; Ignore structure destroy notifies and only
     ;; use substructure destroy notifiers. This way
     ;; event-window is the window's parent.
-    (let ((win (or (find-window window)
-                   (find-withdrawn-window window))))
-      (if win
-          (destroy-window win)
-          (progn
-            (let ((ml (find-mode-line-by-window window)))
-              (when ml (destroy-mode-line ml))))))))
+    (if-let ((win (or (find-window window)
+                      (find-withdrawn-window window))))
+      (destroy-window win)
+      (when-let ((ml (find-mode-line-by-window window)))
+        (destroy-mode-line ml)))))
 
 (defun read-from-keymap (kmaps &optional update-fn)
   "Read a sequence of keys from the user, guided by the keymaps,
@@ -207,9 +209,7 @@ The Caller is responsible for setting up the input focus."
                (when grab (ungrab-pointer)))))
           (match
            (values match key-seq))
-          ((and (find key (list (kbd "?")
-                                (kbd "C-h"))
-                      :test 'equalp))
+          ((and (find key *help-keys* :key #'kbd :test 'equalp))
            (apply 'display-bindings-for-keymaps (reverse (cdr key-seq)) (dereference-kmaps kmaps))
            (values t key-seq))
           (t
@@ -217,41 +217,76 @@ The Caller is responsible for setting up the input focus."
 
 (defun top-maps (&optional (group (current-group)))
   "Return all top level keymaps that are active."
-  (append
-   ;; The plain jane top map is first because that's where users are
-   ;; going to throw in their universally accessible customizations
-   ;; which we don't want groups or minor modes shadowing them.
-   (list '*top-map*)
-   ;; TODO: Minor Mode maps go here
-   ;; lastly, group maps. Last because minor modes should be able to
-   ;; shadow a group's default bindings.
-   (loop for i in *group-top-maps*
-      when (typep group (first i))
-      collect (second i))))
+  (flet ((funcallable-or-kmap (thing)
+           "determine if THING is a function, or if it is bound to a kmap. Used to
+determine whether or not to funcall a symbol or lookup the symbol to retrieve a
+kmap."
+           (or (functionp thing)
+               (and (symbolp thing)
+                    (not (and (boundp thing)
+                              (kmap-p (symbol-value thing))))
+                    (fboundp thing)))))
+    (append
+     ;; The plain jane top map is first because that's where users are
+     ;; going to throw in their universally accessible customizations
+     ;; which we don't want groups or minor modes shadowing them.
+     (list '*top-map*)
+     ;; If a minor mode map element is a function or a symbol that does not denote
+     ;; a keymap but is fbound, then funcall it with the group. Otherwise it
+     ;; should be a kmap object or a symbol bound to a kmap object.
+     (loop for map in *minor-mode-maps*
+           if (funcallable-or-kmap map)
+             append (funcall map group)
+           else
+             collect map)
+     ;; lastly, group maps. Last because minor modes should be able to
+     ;; shadow a group's default bindings.
+     (cond ((typep group 'dynamic-group)
+            (loop for i in *group-top-maps*
+                  when (and (not (eql (first i) 'tile-group))
+                            (typep group (first i)))
+                    collect (second i)))
+           (t (loop for i in *group-top-maps*
+                    when (typep group (first i))
+                      collect (second i)))))))
+
+(defvar *current-key-seq* nil
+  "The sequence of keys which were used to invoke a command, available
+  within a command definition as a dynamic var binding. Commands may
+  dispatch further based on the value in *current-key-seq*. See the
+  REMAP-KEYS contrib module for a working use case.")
+
+(defvar *custom-key-event-handler* nil
+  "A custom key event handler can be set in this variable,
+  which will take precedence over the keymap based handler defined in
+  the default :KEY-PRESS event handler.")
 
 (define-stump-event-handler :key-press (code state #|window|#)
   (labels ((get-cmd (code state)
              (with-focus (screen-key-window (current-screen))
                (handle-keymap (top-maps) code state nil t nil))))
     (unwind-protect
-         ;; modifiers can sneak in with a race condition. so avoid that.
-         (unless (is-modifier code)
-           (multiple-value-bind (cmd key-seq) (get-cmd code state)
-             (cond
-               ((eq cmd t))
-               (cmd
-                (unmap-message-window (current-screen))
-                (eval-command cmd t) t)
-               (t (message "~{~a ~}not bound." (mapcar 'print-key (nreverse key-seq))))))))))
+         (or (and *custom-key-event-handler*
+                  (funcall *custom-key-event-handler* code state))
+             ;; modifiers can sneak in with a race condition. so avoid that.
+             (unless (is-modifier code)
+               (multiple-value-bind (cmd key-seq) (get-cmd code state)
+                 (cond
+                   ((eq cmd t))
+                   (cmd
+                    (unmap-message-window (current-screen))
+                    (let ((*current-key-seq* key-seq))
+                      (eval-command cmd t))
+                    t)
+                   (t (message "~{~a ~}not bound." (mapcar 'print-key (nreverse key-seq)))))))))))
 
 (defun bytes-to-window (bytes)
-  "A sick hack to assemble 4 bytes into a 32 bit number. This is
-because ratpoison sends the rp_command_request window in 8 byte
-chunks."
-  (+ (first bytes)
-     (ash (second bytes) 8)
-     (ash (third bytes) 16)
-     (ash (fourth bytes) 24)))
+  "Combine a list of 4 8-bit bytes into a 32-bit number. This is because
+ratpoison sends the rp_command_request window in 8 byte chunks."
+  (logior (first bytes)
+          (ash (second bytes) 8)
+          (ash (third bytes) 16)
+          (ash (fourth bytes) 24)))
 
 (defun handle-rp-commands (root)
   "Handle a ratpoison style command request."
@@ -274,8 +309,8 @@ chunks."
   "Handle a StumpWM style command request."
   (let* ((win root)
          (screen (find-screen root))
-         (data (xlib:get-property win :stumpwm_command :delete-p t))
-         (cmd (bytes-to-string data)))
+         (data (xlib:get-property win :stumpwm_command :delete-p t :result-type '(vector (unsigned-byte 8))))
+         (cmd (utf8-to-string data)))
     (let ((msgs (screen-last-msg screen))
           (hlts (screen-last-msg-highlights screen))
           (*executing-stumpwm-command* t))
@@ -283,7 +318,7 @@ chunks."
             (screen-last-msg-highlights screen) '())
       (eval-command cmd)
       (xlib:change-property win :stumpwm_command_result
-                            (string-to-bytes (format nil "~{~{~a~%~}~}" (nreverse (screen-last-msg screen))))
+                            (sb-ext:string-to-octets (format nil "~{~{~a~%~}~}" (nreverse (screen-last-msg screen))))
                             :string 8)
       (setf (screen-last-msg screen) msgs
             (screen-last-msg-highlights screen) hlts))
@@ -360,9 +395,8 @@ converted to an atom is removed."
                   screen)
          (handle-stumpwm-commands window))))
     (t
-     (let ((window (find-window window)))
-       (when window
-         (update-window-properties window atom))))))
+     (when-let ((window (find-window window)))
+       (update-window-properties window atom)))))
 
 (define-stump-event-handler :mapping-notify (request start count)
   ;; We could be a bit more intelligent about when to update the
@@ -377,6 +411,20 @@ converted to an atom is removed."
 (define-stump-event-handler :selection-clear (selection)
   (setf (getf *x-selection* selection) nil))
 
+(define-stump-event-handler :selection-notify (window property selection)
+  (dformat 2 "selection-notify: ~s ~s ~s~%" window property selection)
+  (when property
+    (let* ((selection (or selection :primary))
+           (sel-string (utf8-to-string
+                        (xlib:get-property window
+                                           property
+                                           :type :utf8_string
+                                           :result-type 'vector
+                                           :delete-p t))))
+      (when (< 0 (length sel-string))
+        (setf (getf *x-selection* selection) sel-string)
+        (run-hook-with-args *selection-notify-hook* sel-string)))))
+
 (defun find-message-window-screen (win)
   "Return the screen, if any, that message window WIN belongs."
   (dolist (screen *screen-list*)
@@ -457,10 +505,11 @@ converted to an atom is removed."
         (if (eq (window-group window) (current-group))
             (echo-string (window-screen window) (format nil "'~a' denied map request" (window-name window)))
             (echo-string (window-screen window) (format nil "'~a' denied map request in group ~a" (window-name window) (group-name (window-group window))))))
-      (frame-raise-window (window-group window) (window-frame window) window
-                          (if (eq (window-frame window)
-                                  (tile-group-current-frame (window-group window)))
-                              t nil))))
+      (if (typep window 'tile-window)
+          (frame-raise-window (window-group window) (window-frame window) window
+                              (eq (window-frame window)
+                                  (tile-group-current-frame (window-group window))))
+          (raise-window window))))
 
 (defun maybe-raise-window (window)
   (if (deny-request-p window *deny-raise-request*)
@@ -501,37 +550,34 @@ converted to an atom is removed."
              (focus-all our-window)
              (maybe-raise-window our-window)))))
     (:_NET_CLOSE_WINDOW
-     (let ((our-window (find-window window)))
-       (when our-window
-         (delete-window our-window))))
+     (when-let ((our-window (find-window window)))
+       (delete-window our-window)))
     (:_NET_WM_STATE
-     (let ((our-window (find-window window)))
-       (when our-window
-         (let ((action (elt data 0))
-               (p1 (elt data 1))
-               (p2 (elt data 2)))
-           (dolist (p (list p1 p2))
-             ;; Sometimes the number cannot be converted to an atom, so skip them.
-             (unless (or (= p 0)
-                         (not (typep p '(unsigned-byte 29))))
-               (case (safe-atom-name p)
-                 (:_NET_WM_STATE_DEMANDS_ATTENTION
-                  (case action
-                    (1
-                     (add-wm-state window :_NET_WM_STATE_DEMANDS_ATTENTION))
-                    (2
-                     (unless (find-wm-state window :_NET_WM_STATE_DEMANDS_ATTENTION)
-                       (add-wm-state window :_NET_WM_STATE_DEMANDS_ATTENTION))))
-                  (maybe-set-urgency our-window))
-               (:_NET_WM_STATE_FULLSCREEN
-                (update-fullscreen our-window action)))))))))
+     (when-let ((our-window (find-window window))
+                (action (elt data 0))
+                (p1 (elt data 1))
+                (p2 (elt data 2)))
+       (dolist (p (list p1 p2))
+         ;; Sometimes the number cannot be converted to an atom, so skip them.
+         (unless (or (= p 0)
+                     (not (typep p '(unsigned-byte 29))))
+           (case (safe-atom-name p)
+             (:_NET_WM_STATE_DEMANDS_ATTENTION
+              (case action
+                (1
+                 (add-wm-state window :_NET_WM_STATE_DEMANDS_ATTENTION))
+                (2
+                 (unless (find-wm-state window :_NET_WM_STATE_DEMANDS_ATTENTION)
+                   (add-wm-state window :_NET_WM_STATE_DEMANDS_ATTENTION))))
+              (maybe-set-urgency our-window))
+             (:_NET_WM_STATE_FULLSCREEN
+              (update-fullscreen our-window action)))))))
   (:_NET_MOVERESIZE_WINDOW
-   (let ((our-window (find-window window)))
-     (when our-window
-       (let ((x (elt data 1))
-             (y (elt data 2)))
-         (dformat 3 "!!! Data: ~S~%" data)
-         (group-move-request (window-group our-window) our-window x y :root)))))
+   (when-let ((our-window (find-window window))
+              (x (elt data 1))
+              (y (elt data 2)))
+     (dformat 3 "!!! Data: ~S~%" data)
+     (group-move-request (window-group our-window) our-window x y :root)))
   (t
    (dformat 2 "ignored message~%"))))
 
@@ -564,19 +610,63 @@ the window in it's frame."
         (focus-all win)
         (update-all-mode-lines)))))
 
-(define-stump-event-handler :button-press (window code x y child time)
-  (let (screen ml win)
+(defun decode-button-code (code)
+  "Translate the mouse button number into a more readable format"
+  (case code
+    (1 :left-button)
+    (2 :middle-button)
+    (3 :right-button)
+    (4 :wheel-up)
+    (5 :wheel-down)
+    (6 :wheel-left)
+    (7 :wheel-right)
+    (8 :browser-back)
+    (9 :browser-front)
+    (t code)))
+
+(defun scroll-button-keyword-p (button)
+  "Checks if button keyword is generated from the scroll wheel."
+  (or (eq button :wheel-down) (eq button :wheel-up)
+      (eq button :wheel-left) (eq button :wheel-right)))
+
+(defvar *button-state* nil
+  "Modifier state keys for button presses.")
+
+(define-stump-event-handler :button-press (window state code x y child time)
+  (let ((button (decode-button-code code))
+        (screen (find-screen window))
+        (mode-line (find-mode-line-by-window window))
+        (*button-state* (xlib:make-state-keys state))
+        (win))
+    (run-hook-with-args *click-hook* screen code x y)
     (cond
-      ((and (setf screen (find-screen window)) (not child))
-       (group-button-press (screen-current-group screen) x y :root)
-       (run-hook-with-args *root-click-hook* screen code x y))
-      ((setf ml (find-mode-line-by-window window))
-       (run-hook-with-args *mode-line-click-hook* ml code x y))
+      ((setf win (find-window window))
+       ;; We received the event at the parent window, but it occurred in
+       ;; the window itself. We need to adjust the coordinates to make them
+       ;; relative to the parent window.
+       (setf x (+ x (xlib:drawable-x window))
+             y (+ y (xlib:drawable-y window))))
       ((setf win (find-window-by-parent window (top-windows)))
-       (group-button-press (window-group win) x y win))))
+       ;; The click happened on the parent window, so the coordinates are
+       ;; correct.
+       ))
+    (cond
+      ((and screen (not child))
+       (group-button-press (screen-current-group screen) button x y :root)
+       (run-hook-with-args *root-click-hook* screen code x y))
+      (mode-line
+       (run-hook-with-args *mode-line-click-hook* mode-line code x y))
+      (win
+       (group-button-press (window-group win) button x y win))))
   ;; Pass click to client
   (xlib:allow-events *display* :replay-pointer time))
 
+(defun make-xlib-window (drawable)
+  "For some reason the CLX xid cache screws up returns pixmaps when
+they should be windows. So use this function to make a window out of DRAWABLE."
+  (xlib::make-window :id (xlib:drawable-id drawable)
+                     :display *display*))
+
 (defun handle-event (&rest event-slots &key display event-key &allow-other-keys)
   (declare (ignore display))
   (dformat 1 ">>> ~S~%" event-key)
@@ -584,14 +674,14 @@ the window in it's frame."
         (win (getf event-slots :window))
         (*current-event-time* (getf event-slots :time)))
     (when eventfn
-      ;; XXX: In both the clisp and sbcl clx libraries, sometimes what
-      ;; should be a window will be a pixmap instead. In this case, we
-      ;; need to manually translate it to a window to avoid breakage
-      ;; in stumpwm. So far the only slot that seems to be affected is
-      ;; the :window slot for configure-request and reparent-notify
-      ;; events. It appears as though the hash table of XIDs and clx
-      ;; structures gets out of sync with X or perhaps X assigns a
-      ;; duplicate ID for a pixmap and a window.
+      ;; XXX: In sbcl clx libraries, sometimes what should be a window
+      ;; will be a pixmap instead. In this case, we need to manually
+      ;; translate it to a window to avoid breakage in stumpwm. So far
+      ;; the only slot that seems to be affected is the :window slot
+      ;; for configure-request and reparent-notify events. It appears
+      ;; as though the hash table of XIDs and clx structures gets out
+      ;; of sync with X or perhaps X assigns a duplicate ID for a
+      ;; pixmap and a window.
       (when (and win (not (xlib:window-p win)))
         (dformat 10 "Pixmap Workaround! ~s should be a window!~%" win)
         (setf (getf event-slots :window) (make-xlib-window win)))
@@ -606,13 +696,12 @@ the window in it's frame."
               (apply eventfn event-slots))
             (xlib:display-finish-output *display*))
         ((or xlib:window-error xlib:drawable-error) (c)
-          ;; Asynchronous errors are handled in the error
-          ;; handler. Synchronous errors like trying to get the window
-          ;; hints on a deleted window are caught and ignored here. We
-          ;; do this inside the event handler so that the event is
-          ;; handled. If we catch it higher up the event will not be
-          ;; flushed from the queue and we'll get ourselves into an
-          ;; infinite loop.
+          ;; Asynchronous errors are handled in the error handler.
+          ;; Synchronous errors like trying to get the window hints on
+          ;; a deleted window are caught and ignored here. We do this
+          ;; inside the event handler so that the event is handled. If
+          ;; we catch it higher up the event will not be flushed from
+          ;; the queue and we'll get ourselves into an infinite loop.
           (dformat 4 "ignore synchronous ~a~%" c))))
     (dformat 2 "<<< ~S~%" event-key)
     t))
