@@ -97,31 +97,50 @@
   (make-ddump :screens (mapcar 'dump-screen *screen-list*)
               :current (screen-id (current-screen))))
 
+(defun dump-pathname (name)
+  "Convert NAME to a pathname for dump data. If NAME is an absolute path, then it will
+be used as is. Otherwise, defaults to writing to \"FILE.dump\" in the XDG_DATA_HOME 
+location."
+  (if (uiop:absolute-pathname-p name)
+      name
+      (merge-pathnames (ensure-directories-exist (uiop:xdg-data-home #p"stumpwm/"))
+                       (make-pathname :type "dump"
+                                      :name name))))
+
 (defun dump-to-file (foo name)
-  (with-open-file (fp name :direction :output :if-exists :supersede)
+  (with-open-file (fp (dump-pathname name)
+                      :direction :output
+                      :if-exists :supersede
+                      :if-does-not-exist :create)
     (with-standard-io-syntax
       (let ((*package* (find-package :stumpwm))
             (*print-pretty* t))
         (prin1 foo fp)))))
 
-(defcommand dump-group-to-file (file) ((:rest "Dump To File: "))
-  "Dumps the frames of the current group of the current screen to the named file."
+(defcommand dump-group-to-file (file) ((:rest "Dump to file: "))
+  "Dumps the frames of the current group of the current screen to the named file.
+If FILE is an absolute path, then the dump will be read written there.
+Otherwise, defaults to writing to \"FILE.dump\" in the XDG_DATA_HOME location."
   (dump-to-file (dump-group (current-group)) file)
-  (message "Group dumped"))
+  (message "Group dumped."))
 
 (defcommand-alias dump-group dump-group-to-file)
 
-(defcommand dump-screen-to-file (file) ((:rest "Dump To File: "))
-  "Dumps the frames of all groups of the current screen to the named file"
+(defcommand dump-screen-to-file (file) ((:rest "Dump to file: "))
+  "Dumps the frames of all groups of the current screen to the named file.
+If FILE is an absolute path, then the dump will be read written there.
+Otherwise, defaults to writing to \"FILE.dump\" in the XDG_DATA_HOME location."
   (dump-to-file (dump-screen (current-screen)) file)
-  (message "Screen dumped"))
+  (message "Screen dumped."))
 
 (defcommand-alias dump-screen dump-screen-to-file)
 
-(defcommand dump-desktop-to-file (file) ((:rest "Dump To File: "))
-  "Dumps the frames of all groups of all screens to the named file"
+(defcommand dump-desktop-to-file (file) ((:rest "Dump to file: "))
+  "Dumps the frames of all groups of all screens to the named file.
+If FILE is an absolute path, then the dump will be read written there.
+Otherwise, defaults to writing to \"FILE.dump\" in the XDG_DATA_HOME location."
   (dump-to-file (dump-desktop) file)
-  (message "Desktop dumped"))
+  (message "Desktop dumped."))
 
 (defcommand-alias dump-desktop dump-desktop-to-file)
 
@@ -202,9 +221,12 @@
       (when screen
         (restore-screen screen sdump)))))
 
-(defcommand restore-from-file (file) ((:rest "Restore From File: "))
-  "Restores screen, groups, or frames from named file, depending on file's contents."
-  (let ((dump (read-dump-from-file file)))
+(defcommand restore-from-file (file) ((:rest "Restore from file: "))
+  "Restores screen, groups, or frames from named file, depending on file's
+contents. If FILE is an absolute path, then the dump will be read from there.
+Otherwise, defaults to reading from \"FILE.dump\" in the XDG_DATA_HOME location."
+  (let ((dump (read-dump-from-file
+               (dump-pathname file))))
     (typecase dump
       (gdump
        (restore-group (current-group) dump)
@@ -216,10 +238,14 @@
        (restore-desktop dump)
        (message "Desktop restored."))
       (t
-       (message "Don't know how to restore ~a" dump)))))
+       (message "Don't know how to restore ~a." dump)))))
 
 (defcommand-alias restore restore-from-file)
 
 (defcommand place-existing-windows () ()
   "Re-arrange existing windows according to placement rules."
   (sync-window-placement))
+
+(defcommand place-current-window () ()
+  "Re-arrange current window according to placement rules."
+  (sync-single-window-placement (current-screen) (current-window) t))
