@@ -26,22 +26,30 @@
 
 (export '(*top-map*
           *root-map*
+          *key-seq-color*
+          *altgr-offset*
           define-key
-	  kbd
-	  lookup-command
-	  lookup-key
-	  make-sparse-keymap
-	  undefine-key))
+          kbd
+          lookup-command
+          lookup-key
+          make-sparse-keymap
+          undefine-key))
 
 (defvar *top-map* nil
   "The top level key map. This is where you'll find the binding for the
 @dfn{prefix map}.")
 
 (defvar *root-map* nil
-  "This is the keymap by default bound to @kbd{C-t}. It is known as the @dfn{prefix map}.")
+  "This is the keymap by default bound to @kbd{C-t} (along with 
+ *group-root-map* and either *tile-group-root-map*, *float-group-root-map*,
+ or *dynamic-group-map*). It is known as the @dfn{prefix map}.")
+
+(defvar *key-seq-color* "^5"
+  "Color of a keybinding when displayed in windows such as the prefix
+keybinding in the which-key window.")
 
 (defstruct key
-  keysym shift control meta alt hyper super)
+  keysym shift control meta alt hyper super altgr)
 
 (defstruct kmap
   bindings)
@@ -104,10 +112,14 @@ the time these just gets in the way."
     (apply 'xlib:make-state-mask mods)))
 
 (defun report-kbd-parse-error (c stream)
-  (format stream "Failed to parse key string: ~s" (slot-value c 'string)))
+  (format stream "Failed to parse key string: ~s" (slot-value c 'string))
+  (when-let ((reason (kbd-parse-error-reason c)))
+    (format stream "~%Reason: ~A" reason)))
 
 (define-condition kbd-parse-error (stumpwm-error)
-  ((string :initarg :string))
+  ((string :initarg :string)
+   (reason :initarg :reason :reader kbd-parse-error-reason
+	   :initform nil))
   (:report report-kbd-parse-error)
   (:documentation "Raised when a kbd string failed to parse."))
 
@@ -115,29 +127,52 @@ the time these just gets in the way."
   "MODS is a sequence of <MOD CHAR> #\- pairs. Return a list suitable
 for passing as the last argument to (apply #'make-key ...)"
   (unless (evenp end)
-    (signal 'kbd-parse-error :string mods))
-  (apply #'nconc (loop for i from 0 below end by 2
-                       if (char/= (char mods (1+ i)) #\-)
-                       do (signal 'kbd-parse)
-                       collect (case (char mods i)
-                                 (#\M (list :meta t))
-                                 (#\A (list :alt t))
-                                 (#\C (list :control t))
-                                 (#\H (list :hyper t))
-                                 (#\s (list :super t))
-                                 (#\S (list :shift t))
-                                 (t (signal 'kbd-parse-error :string mods))))))
+    (error 'kbd-parse-error :string mods
+           :reason "Did you forget to separate modifier characters with '-'?"))
+  (loop for i from 0 below end by 2
+        when (char/= (char mods (1+ i)) #\-)
+          do (error 'kbd-parse-error :string mods)
+        nconc (case (char mods i)
+                (#\M (list :meta t))
+                (#\A (list :alt t))
+                (#\C (list :control t))
+                (#\H (list :hyper t))
+                (#\s (list :super t))
+                (#\S (list :shift t))
+                (t (error 'kbd-parse-error :string mods
+                          :reason (format nil "Unknown modifer character ~A" (char mods i)))))))
+
+(defvar *altgr-offset* 2
+  "The offset of altgr keysyms. Often 2 or 4, but always an even number.")
+
+(defun keysym-requires-altgr (keysym)
+  (when *display*
+    (unless (and (xlib:keysym->keycodes *display* keysym) t)
+      (let* ((min (xlib:display-min-keycode *display*))
+             (max (xlib:display-max-keycode *display*))
+             (map (xlib::display-keyboard-mapping *display*))
+             (size (array-dimension map 1)))
+        (when (> *altgr-offset* size)
+          (error "AltGr offset is larger than the available offsets"))
+        (do ((i min (1+ i)))
+            ((> i max) nil)
+          (when (or (= keysym (aref map i *altgr-offset*))
+                    (= keysym (aref map i (1+ *altgr-offset*))))
+            (return-from keysym-requires-altgr t)))))))
 
 (defun parse-key (string)
   "Parse STRING and return a key structure. Raise an error of type
 kbd-parse if the key failed to parse."
   (let* ((p (when (> (length string) 2)
               (position #\- string :from-end t :end (- (length string) 1))))
-         (mods (parse-mods string (if p (1+ p) 0)))
-         (keysym (stumpwm-name->keysym (subseq string (if p (1+ p) 0)))))
+         (%mods (parse-mods string (if p (1+ p) 0)))
+         (keysym (stumpwm-name->keysym (subseq string (if p (1+ p) 0))))
+         (mods (if (keysym-requires-altgr keysym)
+                   (append '(:altgr t) %mods)
+                   %mods)))
     (if keysym
         (apply 'make-key :keysym keysym mods)
-        (signal 'kbd-parse-error :string string))))
+        (error 'kbd-parse-error :string string))))
 
 (defun parse-key-seq (keys)
   "KEYS is a key sequence. Parse it and return the list of keys."
@@ -175,12 +210,14 @@ others."
           (keysym->stumpwm-name (key-keysym key))))
 
 (defun print-key-seq (seq)
-  (format nil "^5*~{~a~^ ~}^n" (mapcar 'print-key seq)))
+  (format nil
+          (concat *key-seq-color* "*~{~a~^ ~}^n")
+          (mapcar 'print-key seq)))
 
 (defun define-key (map key command)
   "Add a keybinding mapping for the key, @var{key}, to the command,
 @var{command}, in the specified keymap. If @var{command} is nil, remove an
-exising binding.  For example,
+existing binding.  For example,
 
 @example
 \(stumpwm:define-key stumpwm:*root-map* (stumpwm:kbd \"C-z\") \"echo Zzzzz...\")
