@@ -25,23 +25,25 @@
 
 (in-package :stumpwm)
 
-#+ecl (require "clx")
-
 (export '(*suppress-abort-messages*
           *suppress-frame-indicator*
           *suppress-window-placement-indicator*
           *timeout-wait*
+          *timeout-wait-multiline*
           *timeout-frame-indicator-wait*
           *frame-indicator-text*
           *frame-indicator-timer*
           *message-window-timer*
+          *hooks-enabled-p*
           *command-mode-start-hook*
           *command-mode-end-hook*
           *urgent-window-hook*
           *new-window-hook*
+          *new-head-hook*
           *destroy-window-hook*
           *focus-window-hook*
           *place-window-hook*
+          *pre-thread-hook*
           *start-hook*
           *restart-hook*
           *quit-hook*
@@ -50,6 +52,7 @@
           *focus-frame-hook*
           *new-frame-hook*
           *split-frame-hook*
+          *remove-split-hook*
           *message-hook*
           *top-level-error-hook*
           *focus-group-hook*
@@ -60,6 +63,8 @@
           *mode-line-click-hook*
           *pre-command-hook*
           *post-command-hook*
+          *selection-notify-hook*
+          *menu-selection-hook*
           *display*
           *shell-program*
           *maxsize-border-width*
@@ -69,7 +74,10 @@
           *window-events*
           *window-parent-events*
           *message-window-padding*
+          *message-window-y-padding*
           *message-window-gravity*
+          *message-window-real-gravity*
+          *message-window-input-gravity*
           *editor-bindings*
           *input-window-gravity*
           *normal-gravity*
@@ -124,7 +132,6 @@
           concat
           data-dir-file
           dformat
-          flatten
           define-frame-preference
           redirect-all-output
           remove-hook
@@ -169,15 +176,20 @@
           ;; Conditions
           stumpwm-condition
           stumpwm-error
-          stumpwm-warning))
+          stumpwm-warning
+
+          ;; Completion Options
+          *maximum-completions*
+
+          ;; Minor mode keymaps
+          *minor-mode-maps*))
 
 
-
-(eval-when (:compile-toplevel :load-toplevel :execute)
-  ;; Currently we only support pause-less CALL-IN-MAIN-THREAD for
-  ;; SBCL, since it requires the new io-loop.
-  #+sbcl
-  (pushnew :call-in-main-thread *features*))
+;;; Completions
+(defvar *maximum-completions* 100
+  "Maximum number of completions to show in interactive prompts. Setting
+  this too high can crash the completion process due to drawing too far
+  off screen.")
 
 ;;; Message Timer
 (defvar *suppress-abort-messages* nil
@@ -187,6 +199,10 @@
   "Specifies, in seconds, how long a message will appear for. This must
 be an integer.")
 
+(defvar *timeout-wait-multiline* nil
+  "Specifies, in seconds, how long a message will more than one line will
+appear for. This must be an integer. If falsy, default to *timeout-wait*.")
+
 (defvar *timeout-frame-indicator-wait* 1
   "The amount of time a frame indicator timeout takes.")
 
@@ -262,6 +278,9 @@ arguments: the current window and the last window (could be nil).")
   "A hook called whenever a window is placed by rule. Arguments are
 window group and frame")
 
+(defvar *pre-thread-hook* '()
+  "A hook called before any threads are started. Useful if you need to fork.")
+
 (defvar *start-hook* '()
   "A hook called when stumpwm starts.")
 
@@ -283,13 +302,17 @@ window group and frame")
 called with 2 arguments: the current frame and the last frame.")
 
 (defvar *new-frame-hook* '()
-  "A hook called when a new frame is created. the hook is called with
+  "A hook called when a new frame is created. The hook is called with
 the frame as an argument.")
 
 (defvar *split-frame-hook* '()
   "A hook called when a frame is split. the hook is called with
 the old frame (window is removed), and two new frames as arguments.")
 
+(defvar *remove-split-hook* '()
+  "A hook called when a split is removed. the hook is called with
+the current frame and removed frame as arguments.")
+
 (defvar *message-hook* '()
   "A hook called whenever stumpwm displays a message. The hook
 function is passed any number of arguments. Each argument is a
@@ -313,6 +336,12 @@ sequence it is a part of, and command value bound to the key.")
 window. Called with 4 arguments, the screen containing the root
 window, the button clicked, and the x and y of the pointer.")
 
+(defvar *click-hook* '()
+  "A hook called whenever there is a mouse click.
+Called with 4 arguments, the screen containing the
+window (or nil if there isn't one), the button clicked,
+and the x and y of the pointer.")
+
 (defvar *new-mode-line-hook* '()
   "Called whenever the mode-line is created. It is called with argument,
 the mode-line")
@@ -333,6 +362,18 @@ the command as a symbol.")
   "Called after a command is called. It is called with 1 argument:
 the command as a symbol.")
 
+(defvar *selection-notify-hook* '()
+  "Called after a :selection-notify event is processed. It is called
+with 1 argument: the selection as a string.")
+
+(defvar *menu-selection-hook* '()
+  "Called after an item is selected in the windows menu. It is called
+with 1 argument: the menu.")
+
+(defvar *new-head-hook* '()
+  "A hook called whenever a head is added. It is called with 2 arguments: the
+ new head and the current screen.")
+
 ;; Data types and globals used by stumpwm
 
 (defvar *display* nil
@@ -353,6 +394,9 @@ the command as a symbol.")
 (defvar *text-color* "white"
   "The color of message text.")
 
+(defvar *draw-in-color* t
+  "When NIL color formatters are ignored.")
+
 (defvar *menu-maximum-height* nil
   "Defines the maxium number of lines to display in the menu before enabling
    scrolling. If NIL scrolling is disabled.")
@@ -423,8 +467,11 @@ Include only those we are ready to support.")
 (defvar *message-window-padding* 5
   "The number of pixels that pad the text in the message window.")
 
+(defvar *message-window-y-padding* 0
+  "The number of pixels that pad the text in the message window vertically.")
+
 (defvar *message-window-gravity* :top-right
-  "This variable controls where the message window appears. The follow
+  "This variable controls where the message window appears. The following
 are valid values.
 @table @asis
 @item :top-left
@@ -438,12 +485,27 @@ are valid values.
 @item :bottom
 @end table")
 
+(defvar *message-window-input-gravity* :top-left
+  "This variable controls where the message window appears
+when the input window is being displayed. The following are valid values.
+@table @asis
+@item :top-left
+@item :top-right
+@item :bottom-left
+@item :bottom-right
+@item :center
+@item :top
+@item :left
+@item :right
+@item :bottom
+@end table")
+
 ;; line editor
 (defvar *editor-bindings* nil
   "A list of key-bindings for line editing.")
 
 (defvar *input-window-gravity* :top-right
-  "This variable controls where the input window appears. The follow
+  "This variable controls where the input window appears. The following
 are valid values.
 @table @asis
 @item :top-left
@@ -474,6 +536,7 @@ are valid values.
 (defvar *maxsize-gravity* :center)
 (defvar *transient-gravity* :center)
 
+(declaim (type (member :message :break :abort) *top-level-error-action*))
 (defvar *top-level-error-action* :abort
   "If an error is encountered at the top level, in
 STUMPWM-INTERNAL-LOOP, then this variable decides what action
@@ -499,17 +562,149 @@ Use the window's resource class.
 Use the window's resource name.
 @end table")
 
-(defstruct frame
-  (number nil :type integer)
-  x
-  y
-  width
-  height
-  window)
+(defclass swm-class ()
+  ((new-objects
+    :initform nil
+    :accessor swm-class-new-objects
+    :allocation :class
+    :documentation
+"Track all newly created objects in order to mix in the appropriate minor modes
+when they are touched")))
 
-(defstruct (head (:include frame)))
+(defmethod initialize-instance :after ((obj swm-class) &key &allow-other-keys)
+  ;; Register all newly created objects so that they can have the relevant minor
+  ;; modes autoenabled.
+  (pushnew obj (swm-class-new-objects obj) :test #'eq))
 
-(defclass screen ()
+(defgeneric print-swm-object (object stream)
+  (:method (object stream)
+    (format stream "~A" (type-of object))))
+
+(defmethod print-object ((object swm-class) stream)
+  (print-unreadable-object (object stream)
+    (print-swm-object object stream)
+    (when-let ((minor-modes (list-minor-modes object)))
+      (format stream " :MINOR-MODES ~A" minor-modes))))
+
+(defun make-swm-class-instance (class &rest initargs)
+  "Make an instance of a StumpWM class and autoenable any relevant minor
+modes. CLASS must be a symbol denoting a class which descends, directly or
+indirectly, from swm-class. INITARGS must be all initargs one would pass to
+make-instance."
+  ;; This is implemented as a function instead of as an after method for
+  ;; initialize-instance because autoenabling a minor mode involves changing the
+  ;; class of the object, which is implied to be undefined behavior if called
+  ;; within a method which accesses the objects slots.
+  (declare (special *active-global-minor-modes*))
+  (let ((object (apply #'make-instance class initargs)))
+    (prog1 object
+      (loop for class in *active-global-minor-modes*
+            when (typep object (scope-type (minor-mode-scope class)))
+              do (autoenable-minor-mode class object))
+      (setf (swm-class-new-objects object)
+            (remove object (swm-class-new-objects object) :test #'eq)))))
+
+(defmacro define-swm-class (class-name superclasses slots &rest options)
+  "Define a class and a method for DYNAMIC-MIXINS:REPLACE-CLASS which specializes
+upon the class and replaces it. If SUPERCLASSES is NIL then (SWM-CLASS) is used."
+  (unless superclasses (setq superclasses '(swm-class)))
+  `(progn
+     (defclass ,class-name ,superclasses ,slots ,@options)
+     (defmethod dynamic-mixins:replace-class ((object ,class-name) new &rest r)
+       (apply #'dynamic-mixins:replace-class-in-mixin
+              object new ',class-name r))))
+
+(define-swm-class frame ()
+  ((number
+    :initform nil
+    :initarg :number
+    :accessor frame-number)
+   (x
+    :initform nil
+    :accessor frame-x
+    :initarg :x)
+   (y
+    :initform nil
+    :accessor frame-y
+    :initarg :y)
+   (width
+    :initform nil
+    :accessor frame-width
+    :initarg :width)
+   (height
+    :initform nil
+    :accessor frame-height
+    :initarg :height)
+   (window
+    :initform nil
+    :accessor frame-window
+    :initarg :window)))
+
+(defmethod print-swm-object ((object frame) stream)
+  (format stream "FRAME ~d ~a ~d ~d ~d ~d"
+          (frame-number object) (frame-window object) (frame-x object) (frame-y object) (frame-width object) (frame-height object)))
+
+(defun frame-p (object)
+  (typep object 'frame))
+
+(defun make-frame (&rest rest &key number x y width height window)
+  (declare (ignore number x y width height window))
+  (apply 'make-swm-class-instance 'frame rest))
+
+(defun copy-frame (instance)
+  (make-swm-class-instance 'frame :number (frame-number instance)
+                                  :x (frame-x instance)
+                                  :y (frame-y instance)
+                                  :width (frame-width instance)
+                                  :height (frame-height instance)
+                                  :window (frame-window instance)))
+
+(define-swm-class head (frame)
+  ((name
+    :initform ""
+    :accessor head-name
+    :initarg :name)))
+
+(defmethod print-swm-object ((object head) stream)
+  (write-string "HEAD-" stream)
+  (call-next-method))
+
+;; duplicate frame accessors for heads.
+(macrolet ((define-head-accessor (name)
+             (let ((pkg (find-package :stumpwm)))
+               `(progn
+                  (defgeneric ,(intern (format nil "HEAD-~A" (symbol-name name)) pkg) (head)
+                    (:method ((head head))
+                      (,(intern (format nil "FRAME-~A" (symbol-name name)) pkg) head)))
+                  
+                  (defmethod (setf ,(intern (format nil "HEAD-~A" (symbol-name name)) pkg))
+                      (new (head head))
+                    (setf (,(intern (format nil "FRAME-~A" (symbol-name name)) pkg) head)
+                          new))))))
+  (define-head-accessor number)
+  (define-head-accessor x)
+  (define-head-accessor y)
+  (define-head-accessor width)
+  (define-head-accessor height)
+  (define-head-accessor window))
+
+(defun head-p (object)
+  (typep object 'head))
+
+(defun make-head (&rest rest &key number x y width height window name)
+  (declare (ignore number x y width height window name))
+  (apply 'make-swm-class-instance 'head rest))
+
+(defun copy-head (instance)
+  (make-swm-class-instance 'head :number (frame-number instance)
+                                 :x (frame-x instance)
+                                 :y (frame-y instance)
+                                 :width (frame-width instance)
+                                 :height (frame-height instance)
+                                 :window (frame-window instance)
+                                 :name (head-name instance)))
+
+(define-swm-class screen ()
   ((id :initarg :id :reader screen-id)
    (host :initarg :host :reader screen-host)
    (number :initarg :number :reader screen-number)
@@ -573,17 +768,13 @@ exist, in which case they go into the current group.")
   color-stack
   font)
 
-(defmethod print-object ((object frame) stream)
-  (format stream "#S(frame ~d ~a ~d ~d ~d ~d)"
-          (frame-number object) (frame-window object) (frame-x object) (frame-y object) (frame-width object) (frame-height object)))
-
 (defvar *window-number-map* "0123456789"
   "Set this to a string to remap the window numbers to something more convenient.")
 
 (defvar *group-number-map* "1234567890"
   "Set this to a string to remap the group numbers to something more convenient.")
 
-(defvar *frame-number-map* "0123456789abcdefghijklmnopqrstuvxwyz"
+(defvar *frame-number-map* "0123456789abcdefghijklmnopqrstuvwxyz"
   "Set this to a string to remap the frame numbers to more convenient keys.
 For instance,
 
@@ -617,8 +808,8 @@ char."
 (defvar *modifiers* nil
   "A mapping from modifier type to x11 modifier.")
 
-(defmethod print-object ((object screen) stream)
-  (format stream "#S<screen ~s>" (screen-number object)))
+(defmethod print-swm-object ((object screen) stream)
+  (format stream "SCREEN ~s" (screen-number object)))
 
 (defvar *screen-list* '()
   "The list of screens managed by stumpwm.")
@@ -653,23 +844,26 @@ chosen, resignal the error."
        ,@body)))
 
 ;;; Hook functionality
+(defvar *hooks-enabled-p* t
+  "Controls whether hooks will actually run or not")
 
 (defun run-hook-with-args (hook &rest args)
   "Call each function in HOOK and pass args to it."
-  (handler-case
-      (with-simple-restart (abort-hooks "Abort running the remaining hooks.")
-        (with-restarts-menu
+  (when *hooks-enabled-p*
+    (handler-case
+        (with-simple-restart (abort-hooks "Abort running the remaining hooks.")
+          (with-restarts-menu
             (dolist (fn hook)
               (with-simple-restart (continue-hooks "Continue running the remaining hooks.")
                 (apply fn args)))))
-    (t (c) (message "^B^1*Error on hook ^b~S^B!~% ^n~A" hook c) (values nil c))))
+      (t (c) (message "^B^1*Error on hook ^b~S^B!~% ^n~A" hook c) (values nil c)))))
 
 (defun run-hook (hook)
   "Call each function in HOOK."
   (run-hook-with-args hook))
 
 (defmacro add-hook (hook fn)
-  "Add @var{function} to the hook @var{hook-variable}. For example, to
+  "Add @var{function} to the @var{hook-variable}. For example, to
 display a message whenever you switch frames:
 
 @example
@@ -719,10 +913,10 @@ positive direction."
             min))))
 
 (defun split-seq (seq separators &key test default-value)
-  "split a sequence into sub sequences given the list of seperators."
+  "Split a sequence into subsequences given the list of seperators."
   (let ((seps separators))
     (labels ((sep (c)
-               (find c seps :test test)))
+               (position c seps :test test)))
       (or (loop for i = (position-if (complement #'sep) seq)
                 then (position-if (complement #'sep) seq :start j)
                 as j = (position-if #'sep seq :start (or i 0))
@@ -768,51 +962,6 @@ string which is split to obtain the individual regexps. "
          (post (subseq list nth)))
     (nconc pre (list item) post)))
 
-(defvar *debug-level* 0
-  "Set this variable to a number > 0 to turn on debugging. The greater the number the more debugging output.")
-
-(defvar *debug-expose-events* nil
-  "Set this variable for a visual indication of expose events on internal StumpWM windows.")
-
-(defvar *debug-stream* *error-output*
-  "This is the stream debugging output is sent to. It defaults to
-*error-output*. It may be more convenient for you to pipe debugging
-output directly to a file.")
-
-(defun dformat (level fmt &rest args)
-  (when (>= *debug-level* level)
-    (multiple-value-bind (sec m h) (decode-universal-time (get-universal-time))
-      (format *debug-stream* "~2,'0d:~2,'0d:~2,'0d " h m sec))
-    ;; strip out non base-char chars quick-n-dirty like
-    (write-string (map 'string (lambda (ch)
-                                 (if (typep ch 'standard-char)
-                                     ch #\?))
-                       (apply 'format nil fmt args))
-                  *debug-stream*)
-    (force-output *debug-stream*)))
-
-(defvar *redirect-stream* nil
-  "This variable Keeps track of the stream all output is sent to when
-`redirect-all-output' is called so if it changes we can close it
-before reopening.")
-
-(defun redirect-all-output (file)
-  "Elect to redirect all output to the specified file. For instance,
-if you want everything to go to ~/.stumpwm.d/debug-output.txt you would
-do:
-
-@example
-(redirect-all-output (data-dir-file \"debug-output\" \"txt\"))
-@end example
-"
-  (when (typep *redirect-stream* 'file-stream)
-    (close *redirect-stream*))
-  (setf *redirect-stream* (open file :direction :output :if-exists :append :if-does-not-exist :create)
-        *error-output*    *redirect-stream*
-        *standard-output* *redirect-stream*
-        *trace-output*    *redirect-stream*
-        *debug-stream*    *redirect-stream*))
-
 ;;; 
 ;;; formatting routines
 (defun format-expand (fmt-alist fmt &rest args)
@@ -872,7 +1021,7 @@ with the following formatting options:
 
 @table @asis
 @item %n
-Substitutes the windows number translated via *window-number-map*, if there
+Substitutes the window's number translated via *window-number-map*, if there
 are more windows than *window-number-map* then will use the window-number.
 @item %s
 Substitute the window's status. * means current window, + means last
@@ -892,11 +1041,11 @@ size. For instance, @samp{%20t} crops the window's title to 20
 characters.")
 
 (defvar *window-info-format* "%wx%h %n (%t)"
-  "The format used in the info command.
+  "The format used in the info command. See
   @var{*window-format*} for formatting details.")
 
 (defparameter *window-format-by-class* "%m%n %c %s%50t"
-  "The format used in the info winlist-by-class command.
+  "The format used in the info winlist-by-class command. See
  @var{*window-format*} for formatting details.")
 
 (defvar *group-formatters* '((#\n group-map-number)
@@ -947,10 +1096,10 @@ generally set when killing text in the input bar.")
 recommended this is assigned using LET.")
 
 (defvar *suppress-echo-timeout* nil
-  "Assign this T and messages will not time out. It is recommended this is assigned using LET.")
+  "Assign this T and messages will not time out. It is recommended to assign this using LET.")
 
 (defvar *ignore-echo-timeout* nil
-  "Assign this T and the message time out won't be touched. It is recommended this is assigned using LET.")
+  "Assign this T and the message time out won't be touched. It is recommended to assign this using LET.")
 
 (defvar *run-or-raise-all-groups* t
   "When this is @code{T} the @code{run-or-raise} function searches all groups for a
@@ -989,11 +1138,6 @@ raise/map denial messages will be seen.")
                 deny-list)
        t)))
 
-(defun flatten (list)
-  "Flatten LIST"
-  (labels ( (mklist (x) (if (listp x) x (list x))) )
-    (mapcan #'(lambda (x) (if (atom x) (mklist x) (flatten x))) list)))
-
 (defun list-splice-replace (item list &rest replacements)
   "splice REPLACEMENTS into LIST where ITEM is, removing
 ITEM. Return the new list."
@@ -1070,14 +1214,20 @@ will have no effect.")
   "List of rules governing window placement. Use define-frame-preference to
 add rules")
 
-(defmacro define-frame-preference (target-group &rest frame-rules)
+(defmacro define-frame-preference (target-group &body frame-rules)
   "Create a rule that matches windows and automatically places them in
 a specified group and frame. Each frame rule is a lambda list:
 @example
-\(frame-number raise lock &key create restore dump-name class instance type role title)
+\(frame-number raise lock &key from-group create restore dump-name class class-not
+instance instance-not type type-not role role-not title title-not
+match-properties-and-function match-properties-or-function)
 @end example
 
 @table @var
+@item target-group
+When nil, rule applies in the current group. When non nil, @var{lock} determines
+applicability of rule
+
 @item frame-number
 The frame number to send matching windows to
 
@@ -1085,12 +1235,19 @@ The frame number to send matching windows to
 When non-nil, raise and focus the window in its frame
 
 @item lock
-When this is nil, this rule will only match when the current group
-matches @var{target-group}. When non-nil, this rule matches regardless
+When this is nil, this rule will only match when @var{target-group}
+matches the group designated by @var{from-group}.
+When non-nil, this rule matches regardless
 of the group and the window is sent to @var{target-group}. If
 @var{lock} and @var{raise} are both non-nil, then stumpwm will jump to
 the specified group and focus the matched window.
 
+@item from-group
+When @var{lock} is NIL, and this is non-NIL, this rule will only match
+when @var{target-group} matches @var{from-group}. This should be set
+to either a group name(a string), or an expression that returns a group(e.g (current-group)).
+When this is NIL, the rule matches if @var{target-group} matches
+the group the window is in, or the current group if the window has no group.
 @item create
 When non-NIL the group is created and eventually restored when the value of
 create is a group dump filename in *DATA-DIR*. Defaults to NIL.
@@ -1100,27 +1257,49 @@ When non-NIL the group is restored even if it already exists. This arg should
 be set to the dump filename to use for forced restore. Defaults to NIL
 
 @item class
-The window's class must match @var{class}.
+The windows class must match @var{class}.
+
+@item class-not
+The windows class must not match @var{class-not}
 
 @item instance
-The window's instance/resource name must match @var{instance}.
+The windows instance/resource name must match @var{instance}.
+
+@item instance-not
+The windows instance/resource name must not match @var{instance-not}.
 
 @item type
-The window's type must match @var{type}.
+The windows type must match @var{type}.
+
+@item type-not
+The windows type must not match @var{type-not}.
 
 @item role
-The window's role must match @var{role}.
+The windows role must match @var{role}.
+
+@item role-not
+The windows role must not match @var{role-not}.
 
 @item title
-The window's title must match @var{title}.
+The windows title must match @var{title}.
+
+@item title-not
+The windows title must not match @var{title-not}.
+
+@item match-properties-and-function
+A function that, if provided, must return true alongside the provided properties
+in order for the rule to match. This function takes one argument, the window. 
+Must be an unquoted symbol to be looked up at runtime. 
+
+@item match-properties-or-function
+A function that, if provided and returning true, will cause the rule to match
+regardless of whether the window properties match. Takes one argument, the window.
+Must be an unquoted symbol to be looked up at runtime. 
 @end table"
   (let ((x (gensym "X")))
     `(dolist (,x ',frame-rules)
        ;; verify the correct structure
-       (destructuring-bind (frame-number raise lock
-                                         &rest keys
-                                         &key create restore class instance type role title) ,x
-         (declare (ignore create restore class instance type role title))
+       (destructuring-bind (frame-number raise lock &rest keys) ,x
          (push (list* ,target-group frame-number raise lock keys)
                *window-placement-rules*)))))
 
@@ -1128,12 +1307,55 @@ The window's title must match @var{title}.
   "Clear all window placement rules."
   (setf *window-placement-rules* nil))
 
+(defvar *fullscreen-in-frame-p-window-functions* nil
+  "A alist of predicate functions for determining if a window should be
+fullscreen in frame.")
+
+(defun fullscreen-in-frame-p (win)
+  (some (lambda (r)
+          (let ((res (funcall (cdr r) win)))
+            (when res
+              (dformat 3 "Fullscreen in frame selector ~A matches window ~A"
+                       (car r) win))
+            res))
+        *fullscreen-in-frame-p-window-functions*))
+
+(defun add-fullscreen-in-frame-rule (name function &key shadow)
+  "Add a function to the fullscreen-in-frame window rules alist.  If @var{NAME}
+already exists as a key in the alist and @var{SHADOW} is nil, then
+@var{FUNCTION} replaces the existing value.  Otherwise @var{NAME} and
+@var{FUNCTION} are pushed onto the alist."
+  (let ((present (assoc name *fullscreen-in-frame-p-window-functions*)))
+    (if (and present (not shadow))
+        (setf (cdr present) function)
+        (push (cons name function) *fullscreen-in-frame-p-window-functions*))))
+
+(defun remove-fullscreen-in-frame-rule (name &key count)
+  "Remove rules named @var{NAME} from the fullscreen-in-frame window rules alist.
+If @var{COUNT} is NIL then all matching rules are removed, otherwise only the
+first @var{COUNT} rules are removed."
+  (setf *fullscreen-in-frame-p-window-functions*
+        (remove name *fullscreen-in-frame-p-window-functions*
+                :key #'car :count count)))
+
+(defmacro define-fullscreen-in-frame-rule (name (window-argument) &body body)
+  "Define a rule for a window to be fullscreened within the frame.  Each rule is a
+function which will be called when a window is made fullscreen.  If the rule
+returns NIL then the fullscreen window takes up the entire head, otherwise it
+takes up only its frame. Within the body of the rule @var{WINDOW-ARGUMENT} is
+bound to the window being processed."
+  `(flet ((,name (,window-argument) ,@body))
+     (add-fullscreen-in-frame-rule ',name #',name)))
+
 (defvar *mouse-focus-policy* :ignore
   "The mouse focus policy decides how the mouse affects input
 focus. Possible values are :ignore, :sloppy, and :click. :ignore means
 stumpwm ignores the mouse. :sloppy means input focus follows the
 mouse; the window that the mouse is in gets the focus. :click means
-input focus is transfered to the window you click on.")
+input focus is transfered to the window you click on.
+
+If *MOUSE-FOCUS-POLICY* holds any value other than those listed above,
+mouse focus will behave as though it contains :IGNORE")
 
 (defvar *root-click-focuses-frame* t
   "Set to NIL if you don't want clicking the root window to focus the frame
@@ -1249,3 +1471,29 @@ of :error."
 
 (defun command-mode-end-message ()
   (message "Exited command-mode."))
+
+(defstruct (mode-line (:constructor %make-mode-line))
+  screen
+  head
+  window
+  format
+  position
+  contents
+  cc
+  height
+  factor
+  (mode :stump)
+  on-click-bounds
+  new-bounds)
+
+(defstruct timer
+  time repeat function args)
+
+(defvar *minor-mode-maps* ()
+  "A list of minor mode keymaps. An element of the list may be a single keymap or
+a function. If an element is a function it must take a group instance and return
+a list of keymaps.")
+
+(defvar *custom-command-filters* ()
+  "A list of functions which take a group instance and a command structure, and
+return true when the command should be active.")
