@@ -24,14 +24,15 @@
 
 (in-package #:stumpwm)
 
-(export '(current-screen
+(export '(*default-bg-color*
+          current-screen
           current-window
           screen-current-window
-	  screen-number
-	  screen-groups
-	  screen-windows
-	  screen-height
-	  screen-width
+          screen-number
+          screen-groups
+          screen-windows
+          screen-height
+          screen-width
           set-fg-color
           set-bg-color
           set-border-color
@@ -44,6 +45,9 @@
           set-frame-outline-width
           set-font))
 
+(defvar *default-bg-color* #x333333
+  "Default color for the desktop background.")
+
 ;; Screen helper functions
 
 (defun translate-id (src src-start src-end font dst dst-start)
@@ -220,7 +224,7 @@ identity with a range check."
   (update-screen-color-context screen))
 
 (defun update-colors-all-screens ()
-  "After setting the fg, bg, or border colors. call this to sync any existing windows."
+  "After setting the fg, bg, or border colors, call this to sync any existing windows."
   (mapc 'update-colors-for-screen *screen-list*))
 
 (defun update-border-for-screen (screen)
@@ -383,7 +387,11 @@ FOCUS-WINDOW is an extra window used for _NET_SUPPORTING_WM_CHECK."
   (xlib:display-finish-output *display*)
   ;; Initialize the screen structure
   (labels ((ac (color)
-             (xlib:alloc-color (xlib:screen-default-colormap screen-number) color)))
+             ;; We add an alpha channel to the color returned by
+             ;; xlib:alloc-color. This is normally done by stumpwm:alloc-color,
+             ;; but that requires a screen instance.
+             (logior (xlib:alloc-color (xlib:screen-default-colormap screen-number) color)
+                     (ash #xff 24))))
     (let* ((default-colormap (xlib:screen-default-colormap screen-number))
            (screen-root (xlib:screen-root screen-number))
            (fg-color (ac +default-foreground-color+))
@@ -395,9 +403,12 @@ FOCUS-WINDOW is an extra window used for _NET_SUPPORTING_WM_CHECK."
            (float-focus-color (ac +default-float-focus-color+))
            (float-unfocus-color (ac +default-float-unfocus-color+))
            (font (open-font *display*
-                            (if (font-exists-p +default-font-name+)
-                                +default-font-name+
-                                "*")))
+                            (cond ((font-exists-p +default-font-name+)
+                                   +default-font-name+)
+                                  ((font-exists-p "fixed")
+                                   "fixed")
+                                  (t
+                                   "*"))))
            (message-window (xlib:create-window :parent screen-root
                                                :x 0 :y 0 :width 1 :height 1
                                                :colormap default-colormap
@@ -406,69 +417,71 @@ FOCUS-WINDOW is an extra window used for _NET_SUPPORTING_WM_CHECK."
                                                :border-width 1
                                                :bit-gravity :north-east
                                                :event-mask '(:exposure)))
-           (screen (make-instance 'screen
-                                  :id id
-                                  :host host
-                                  :number screen-number
-                                  :border-color border-color
-                                  :fg-color fg-color
-                                  :bg-color bg-color
-                                  :win-bg-color win-bg-color
-                                  :focus-color focus-color
-                                  :unfocus-color unfocus-color
-                                  :float-focus-color float-focus-color
-                                  :float-unfocus-color float-unfocus-color
-                                  :msg-border-width 1
-                                  :frame-outline-width +default-frame-outline-width+
-                                  :fonts (list font)
-                                  :input-window (xlib:create-window
-                                                 :parent screen-root
-                                                 :x 0 :y 0 :width 20 :height 20
-                                                 :colormap default-colormap
-                                                 :background bg-color
-                                                 :border border-color
-                                                 :border-width 1
-                                                 :event-mask '(:key-press :key-release))
-                                  :focus-window (xlib:create-window
-                                                 :parent screen-root
-                                                 :x 0 :y 0 :width 1 :height 1)
-                                  :key-window (xlib:create-window
-                                               :parent screen-root
-                                               :x 0 :y 0 :width 1 :height 1
-                                               :event-mask '(:key-press :key-release))
-                                  :frame-window (xlib:create-window
-                                                 :parent screen-root
-                                                 :x 0 :y 0 :width 20 :height 20
-                                                 :colormap default-colormap
-                                                 :background bg-color
-                                                 :border border-color
-                                                 :border-width 1
-                                                 :event-mask '(:exposure))
-                                  :frame-outline-gc (xlib:create-gcontext
-                                                     :drawable screen-root
-                                                     :font (when (typep font 'xlib:font) font)
-                                                     :foreground fg-color
-                                                     :background fg-color
-                                                     :line-style :double-dash
-                                                     :line-width +default-frame-outline-width+)
-                                  :message-cc (make-ccontext
-                                               :win message-window
-                                               :font font
-                                               :gc (xlib:create-gcontext
-                                                    :drawable message-window
-                                                    :font (when (typep font 'xlib:font) font)
-                                                    :foreground fg-color
-                                                    :background bg-color))))
-           (group (make-instance 'tile-group
-                                 :screen screen
-                                 :number 1
-                                 :name *default-group-name*)))
+           (screen (make-swm-class-instance
+                    'screen
+                    :id id
+                    :host host
+                    :number screen-number
+                    :border-color border-color
+                    :fg-color fg-color
+                    :bg-color bg-color
+                    :win-bg-color win-bg-color
+                    :focus-color focus-color
+                    :unfocus-color unfocus-color
+                    :float-focus-color float-focus-color
+                    :float-unfocus-color float-unfocus-color
+                    :msg-border-width 1
+                    :frame-outline-width +default-frame-outline-width+
+                    :fonts (list font)
+                    :input-window (xlib:create-window
+                                   :parent screen-root
+                                   :x 0 :y 0 :width 20 :height 20
+                                   :colormap default-colormap
+                                   :background bg-color
+                                   :border border-color
+                                   :border-width 1
+                                   :event-mask '(:key-press :key-release))
+                    :focus-window (xlib:create-window
+                                   :parent screen-root
+                                   :x 0 :y 0 :width 1 :height 1)
+                    :key-window (xlib:create-window
+                                 :parent screen-root
+                                 :x 0 :y 0 :width 1 :height 1
+                                 :event-mask '(:key-press :key-release))
+                    :frame-window (xlib:create-window
+                                   :parent screen-root
+                                   :x 0 :y 0 :width 20 :height 20
+                                   :colormap default-colormap
+                                   :background bg-color
+                                   :border border-color
+                                   :border-width 1
+                                   :event-mask '(:exposure))
+                    :frame-outline-gc (xlib:create-gcontext
+                                       :drawable screen-root
+                                       :font (when (typep font 'xlib:font) font)
+                                       :foreground fg-color
+                                       :background fg-color
+                                       :line-style :double-dash
+                                       :line-width +default-frame-outline-width+)
+                    :message-cc (make-ccontext
+                                 :win message-window
+                                 :font font
+                                 :gc (xlib:create-gcontext
+                                      :drawable message-window
+                                      :font (when (typep font 'xlib:font) font)
+                                      :foreground fg-color
+                                      :background bg-color))))
+           (group (make-swm-class-instance 'tile-group
+                                           :screen screen
+                                           :number 1
+                                           :name *default-group-name*)))
       (setf (screen-groups screen) (list group)
             (screen-current-group screen) group
             (ccontext-screen (screen-message-cc screen)) screen
             (screen-heads screen) (make-screen-heads screen screen-root)
             (tile-group-frame-tree group) (copy-heads screen)
-            (tile-group-current-frame group) (first (tile-group-frame-tree group)))
+            (tile-group-current-frame group) (first (tile-group-frame-tree group))
+            (xlib:window-background screen-root) *default-bg-color*)
       ;; The focus window is mapped at all times
       (xlib:map-window (screen-focus-window screen))
       (xlib:map-window (screen-key-window screen))
