@@ -22,12 +22,15 @@
 
 (in-package :stumpwm)
 
-(export '(cancel-timer
-	  run-with-timer
-          *toplevel-io*
-	  stumpwm
-	  timer-p))
+(export '(*toplevel-io*
+          stumpwm
+          call-in-main-thread
+          in-main-thread-p
+          push-event
+          close-resources))
 
+(defvar *in-main-thread* nil
+  "Dynamically bound to T during the execution of the main stumpwm function.")
 
 ;;; Main
 
@@ -37,17 +40,12 @@ doesn't exist. Returns a values list: whether the file loaded (t if no
 rc files exist), the error if it didn't, and the rc file that was
 loaded. When CATCH-ERRORS is nil, errors are left to be handled
 further up. "
-  (let* ((xdg-config-dir
-           (let ((dir (getenv "XDG_CONFIG_HOME")))
-             (if (or (not dir) (string= dir ""))
-                 (merge-pathnames  #p".config/" (user-homedir-pathname))
-                 dir)))
-         (user-rc
+  (let* ((user-rc
            (probe-file (merge-pathnames #p".stumpwmrc" (user-homedir-pathname))))
          (dir-rc
            (probe-file (merge-pathnames #p".stumpwm.d/init.lisp" (user-homedir-pathname))))
          (conf-rc
-           (probe-file (merge-pathnames #p"stumpwm/config" xdg-config-dir)))
+           (probe-file (uiop:xdg-config-home #p"stumpwm/config/")))
          (etc-rc (probe-file #p"/etc/stumpwmrc"))
          (rc (or user-rc dir-rc conf-rc etc-rc)))
     (if rc
@@ -62,7 +60,7 @@ further up. "
 
 (defun error-handler (display error-key &rest key-vals &key asynchronous &allow-other-keys)
   "Handle X errors"
-  (cond 
+  (cond
     ;; ignore asynchronous window errors
     ((and asynchronous
           (find error-key '(xlib:window-error xlib:drawable-error xlib:match-error)))
@@ -72,99 +70,10 @@ further up. "
      (throw :top-level :quit))
      ;; all other asynchronous errors are printed.
      (asynchronous
-      (message "Caught Asynchronous X Error: ~s ~s" error-key key-vals))
+      (message "Caught Asynchronous X Error: ~s ~s." error-key key-vals))
      (t
       (apply 'error error-key :display display :error-key error-key key-vals))))
 
-;;; Timers
-
-(defvar *toplevel-io* nil
-  "Top-level I/O loop")
-
-(defvar *timer-list* nil
-  "List of active timers.")
-
-(defvar *timer-list-lock* (make-lock)
-  "Lock that should be held whenever *TIMER-LIST* is modified.")
-
-(defvar *in-main-thread* nil
-  "Dynamically bound to T during the execution of the main stumpwm function.")
-
-(defstruct timer
-  time repeat function args)
-
-(defun run-with-timer (secs repeat function &rest args)
-  "Perform an action after a delay of SECS seconds.
-Repeat the action every REPEAT seconds, if repeat is non-nil.
-SECS and REPEAT may be reals.
-The action is to call FUNCTION with arguments ARGS."
-  (check-type secs (real 0 *))
-  (check-type repeat (or null (real 0 *)))
-  (check-type function (or function symbol))
-  (let ((timer (make-timer
-                :repeat repeat
-                :function function
-                :args args)))
-    (schedule-timer timer secs)
-    (labels ((append-to-list ()
-               (with-lock-held (*timer-list-lock*)
-                 (setf *timer-list* (merge 'list *timer-list* (list timer) #'< :key #'timer-time)))))
-      ;; If CALL-IN-MAIN-THREAD is supported, the timer should be scheduled in the main thread.
-      #+call-in-main-thread
-      (call-in-main-thread #'append-to-list)
-      #-call-in-main-thread
-      (append-to-list)
-      timer)))
-
-(defun cancel-timer (timer)
-  "Remove TIMER from the list of active timers."
-  (check-type timer timer)
-  (with-lock-held (*timer-list-lock*)
-    (setf *timer-list* (remove timer *timer-list*))))
-
-(defun schedule-timer (timer when)
-  (setf (timer-time timer) (+ (get-internal-real-time)
-                              (* when internal-time-units-per-second))))
-
-(defun sort-timers (timers)
-  (let ((now (get-internal-real-time))
-        (pending ())
-        (remaining ()))
-    (dolist (timer timers)
-      (if (<= (timer-time timer) now)
-          (progn (push timer pending)
-                 (when (timer-repeat timer)
-                   (schedule-timer timer (timer-repeat timer))
-                   (push timer remaining)))
-          (push timer remaining)))
-    (values pending remaining)))
-
-(defun update-timer-list (timers)
-  "Update the timer list, sorting the timers by which is closer expiry."
-  (setf *timer-list*
-        (sort timers #'< :key #'timer-time)))
-
-(defun execute-timers (timers)
-  (map nil #'execute-timer timers))
-
-(defun execute-timer (timer)
-  (apply (timer-function timer) (timer-args timer)))
-
-(defun run-expired-timers ()
-  (let ((expired (with-lock-held (*timer-list-lock*)
-                   (multiple-value-bind (pending remaining)
-                       (sort-timers *timer-list*)
-                     (update-timer-list remaining)
-                     pending))))
-    ;; Call the timers after the lock has been released
-    (execute-timers expired)))
-
-(defun get-next-timeout (timers)
-  "Return the number of seconds until the next timeout or nil if there are no timers."
-  (when timers
-    (max (/ (- (timer-time (car timers)) (get-internal-real-time))
-            internal-time-units-per-second)
-         0)))
 
 (defgeneric handle-top-level-condition (c))
 
@@ -179,7 +88,9 @@ The action is to call FUNCTION with arguments ARGS."
 (defmethod handle-top-level-condition ((c serious-condition))
   (when (and (find-restart :remove-channel)
              (not (typep *current-io-channel*
-                         '(or stumpwm-timer-channel display-channel request-channel))))
+                         '(or stumpwm-timer-channel
+                           display-channel
+                           request-channel))))
     (message "Removed channel ~S due to uncaught error '~A'." *current-io-channel* c)
     (invoke-restart :remove-channel))
   (ecase *top-level-error-action*
@@ -188,24 +99,14 @@ The action is to call FUNCTION with arguments ARGS."
        (write-line s)
        (print-backtrace)
        (message "^1*^B~a" s)))
-    (:break (invoke-debugger c))
+    (:break (restart-case
+                (invoke-debugger c)
+              (:abort-debugging ()
+                :report (lambda (stream) (format stream "abort debugging"))
+                (throw :top-level (list c (backtrace-string))))))
     (:abort
      (throw :top-level (list c (backtrace-string))))))
 
-(defclass stumpwm-timer-channel () ())
-
-(defmethod io-channel-ioport (io-loop (channel stumpwm-timer-channel))
-  (declare (ignore io-loop))
-  nil)
-(defmethod io-channel-events ((channel stumpwm-timer-channel))
-  (with-lock-held (*timer-list-lock*)
-    (if *timer-list*
-        `((:timeout ,(timer-time (car *timer-list*))))
-        '(:loop))))
-(defmethod io-channel-handle ((channel stumpwm-timer-channel) (event (eql :timeout)) &key)
-  (run-expired-timers))
-(defmethod io-channel-handle ((channel stumpwm-timer-channel) (event (eql :loop)) &key)
-  (run-expired-timers))
 
 (defclass request-channel ()
   ((in    :initarg :in
@@ -214,7 +115,7 @@ The action is to call FUNCTION with arguments ARGS."
           :reader request-channel-out)
    (queue :initform nil
           :accessor request-channel-queue)
-   (lock  :initform (make-lock)
+   (lock  :initform (sb-thread:make-mutex)
           :reader request-channel-lock)))
 
 (defvar *request-channel* nil)
@@ -233,38 +134,32 @@ The action is to call FUNCTION with arguments ARGS."
     with in = (request-channel-in channel)
     do (read-byte in)
     while (listen in))
-  (let ((events (with-lock-held ((request-channel-lock channel))
+  (let ((events (sb-thread:with-mutex ((request-channel-lock channel))
                   (let ((queue-copy (request-channel-queue channel)))
                     (setf (request-channel-queue channel) nil)
                     queue-copy))))
     (dolist (event (reverse events))
       (funcall event))))
 
-#+ccl
-(defmethod io-channel-ioport (io-loop (channel ccl::fd-binary-input-stream))
-  (declare (ignore io-loop))
-  (ccl::stream-device channel :input))
+(defun in-main-thread-p ()
+  *in-main-thread*)
+
+(defun push-event (fn)
+  (sb-thread:with-mutex ((request-channel-lock *request-channel*))
+    (push fn (request-channel-queue *request-channel*)))
+  (let ((out (request-channel-out *request-channel*)))
+    ;; For now, just write a single byte since all we want is for the
+    ;; main thread to process the queue. If we want to handle
+    ;; different types of events, we'll have to change this so that
+    ;; the message sent indicates the event type instead.
+    (write-byte 0 out)
+    (finish-output out)))
 
-#+call-in-main-thread
 (defun call-in-main-thread (fn)
-  (cond (*in-main-thread*
+  (cond ((in-main-thread-p)
          (funcall fn))
         (t
-         (with-lock-held ((request-channel-lock *request-channel*))
-           (push fn (request-channel-queue *request-channel*)))
-         (let ((out (request-channel-out *request-channel*)))
-           ;; For now, just write a single byte since all we want is for the
-           ;; main thread to process the queue. If we want to handle
-           ;; different types of events, we'll have to change this so that
-           ;; the message sent indicates the event type instead.
-           (write-byte 0 out)
-           (finish-output out)))))
-
-#-call-in-main-thread
-(defun call-in-main-thread (fn)
-  (if *in-main-thread*
-      (funcall fn)
-      (run-with-timer 0 nil fn)))
+         (push-event fn))))
 
 (defclass display-channel ()
   ((display :initarg :display)))
@@ -299,7 +194,6 @@ The action is to call FUNCTION with arguments ARGS."
 
          ;; If we have no implementation for the current CL, then
          ;; don't register the channel.
-         #+call-in-main-thread
          (multiple-value-bind (in out)
              (open-pipe)
            (let ((channel (make-instance 'request-channel :in in :out out)))
@@ -316,18 +210,28 @@ The action is to call FUNCTION with arguments ARGS."
 (defun parse-display-string (display)
   "Parse an X11 DISPLAY string and return the host and display from it."
   (ppcre:register-groups-bind (protocol host ('parse-integer display screen))
-			      ("^(?:(.*?)/)?(.*?)?:(\\d+)(?:\\.(\\d+))?" display :sharedp t)
-    (values 
+                              ("^(?:(.*?)/)?(.*?)?:(\\d+)(?:\\.(\\d+))?" display :sharedp t)
+    (values
      ;; clx doesn't like (vector character *)
      (coerce (or host "")
-	     '(simple-array character (*)))
+             '(simple-array character (*)))
      display screen
      (cond (protocol
-	    (intern1 protocol :keyword))
-	   ((or (string= host "")
-		(string-equal host "unix"))
-	    :local)
-	   (t :internet)))))
+             (intern1 protocol :keyword))
+           ((or (string= host "")
+                (string-equal host "unix"))
+            :local)
+           (t :internet)))))
+
+(defun ensure-data-dir ()
+  (ensure-directories-exist (data-dir) :mode #o700))
+
+(defun data-dir ()
+  (merge-pathnames ".stumpwm.d/" (user-homedir-pathname)))
+
+(defun close-resources ()
+  (xlib:close-display *display*)
+  (close-log))
 
 (defun stumpwm-internal (display-str)
   (multiple-value-bind (host display screen protocol) (parse-display-string display-str)
@@ -339,21 +243,28 @@ The action is to call FUNCTION with arguments ARGS."
       (unwind-protect
            (progn
              (let ((*initializing* t))
-               ;; we need to do this first because init-screen grabs keys
+               (ensure-data-dir)
+               (open-log)
+               
+               ;; we need to do this first because init-screen grabs
+               ;; keys
                (update-modifier-map)
                ;; Initialize all the screens
                (setf *screen-list* (loop for i in (xlib:display-roots *display*)
                                       for n from 0
                                       collect (init-screen i n host)))
                (xlib:display-finish-output *display*)
+               ;; Enable minor mode keymap lookup. This needs to be done after
+               ;; screens are initialized.
+               (push #'minor-mode-top-maps *minor-mode-maps*)
                ;; Load rc file
                (let ((*package* (find-package *default-package*)))
                  (multiple-value-bind (success err rc) (load-rc-file)
                    (if success
                        (and *startup-message* (message *startup-message* (print-key *escape-key*)))
-                       (message "^B^1*Error loading ^b~A^B: ^n~A" rc err))))
+                       (message "^B^1*Error loading ^b~A^B: ^n~A." rc err))))
                (when *last-unhandled-error*
-                 (message-no-timeout "^B^1*StumpWM Crashed With An Unhandled Error!~%Copy the error to the clipboard with the 'copy-unhandled-error' command.~%^b~a^B^n~%~%~a"
+                 (message-no-timeout "^B^1*StumpWM Crashed With An Unhandled Error!~%Copy the error to the clipboard with the 'copy-unhandled-error' command.~%^b~a^B^n~%~%~a."
                           (first *last-unhandled-error*) (second *last-unhandled-error*)))
                (mapc 'process-existing-windows *screen-list*)
                ;; We need to setup each screen with its current window. Go
@@ -371,17 +282,45 @@ The action is to call FUNCTION with arguments ARGS."
                    (when (and netwm-id (< netwm-id (length (screen-groups s))))
                      (switch-to-group (elt (sort-groups s) netwm-id))))
                  (redraw-current-message (current-screen))))
+
+             (run-hook *pre-thread-hook*)
+
+             ;; Start hashing the user's PATH so completion is quick
+             ;; the first time they try to run a command.
+             (sb-thread:make-thread #'rehash)
+
              ;; Let's manage.
              (let ((*package* (find-package *default-package*)))
                (run-hook *start-hook*)
                (stumpwm-internal-loop)))
-        (xlib:close-display *display*))))
+        (close-resources))))
   ;; what should the top level loop do?
   :quit)
 
+(defun force-stumpwm-restart (&key (close-display t))
+  (when close-display
+    (xlib:close-display *display*))
+  (apply 'execv (first sb-ext:*posix-argv*) sb-ext:*posix-argv*))
+
+;; based on cffi version of set-signal-handler from Andrew Lyon at https://stackoverflow.com/a/10442062
+;; rewritten to use SBCL's Foreign Function Interface directly by Max-Gerd Retzlaff
+(defmacro set-signal-handler (signo &body body)
+  `(sb-alien:alien-funcall
+    (sb-alien:extern-alien "signal" (function sb-alien:void
+                                              sb-alien:int sb-alien:system-area-pointer))
+    ,signo
+    ;; callback function
+    (sb-alien:alien-sap
+     (sb-alien::alien-lambda sb-alien:void ((signum sb-alien:int))
+       (declare (ignore signum))
+       ,@body))))
+
 ;; Usage: (stumpwm)
 (defun stumpwm (&optional (display-str (or (getenv "DISPLAY") ":0")))
   "Start the stump window manager."
+  (set-signal-handler sb-posix:sighup
+    (dformat 0 "SIGHUP received: forcing immediate restart of stumpwm~%") ;; debug level 0 to "force" logging
+    (force-stumpwm-restart))
   (let ((*in-main-thread* t))
     (setf *data-dir*
           (make-pathname :directory (append (pathname-directory (user-homedir-pathname))
@@ -393,14 +332,14 @@ The action is to call FUNCTION with arguments ARGS."
         (setf *last-unhandled-error* nil)
         (cond ((and (consp ret)
                     (typep (first ret) 'condition))
-               (format t "~&Caught '~a' at the top level. Please report this.~%~a" 
+               (format t "~&Caught '~a' at the top level. Please report this.~%~a"
                        (first ret) (second ret))
                (setf *last-unhandled-error* ret))
               ;; we need to jump out of the event loop in order to hup
               ;; the process because otherwise we get errors.
               ((eq ret :hup-process)
                (run-hook *restart-hook*)
-               (apply 'execv (first (argv)) (argv)))
+               (force-stumpwm-restart :close-display nil))
               ((eq ret :restart)
                (run-hook *restart-hook*))
               (t
