@@ -8,16 +8,19 @@
   "Set this to T if you never want windows to resize based on incremental WM_HINTs,
 like xterm and emacs.")
 
-(defclass tile-window (window)
-  ((frame   :initarg :frame   :accessor window-frame)
+(define-swm-class tile-window (window)
+  ((frame   :initarg :frame   :accessor window-frame :type frame)
    (normal-size :initform nil :accessor window-normal-size)))
 
+(defmethod print-swm-object ((object tile-window) stream)
+  (write-string "TILE-" stream)
+  (call-next-method))
+
 (defmethod update-decoration ((window tile-window))
   ;; give it a colored border but only if there are more than 1 frames.
   (let* ((group (window-group window))
          (screen (group-screen group)))
-    (let ((c (if (and (> (length (group-frames group)) 1)
-                      (eq (group-current-window group) window))
+    (let ((c (if (eq (group-current-window group) window)
                  (screen-focus-color screen)
                  (screen-unfocus-color screen))))
       (setf (xlib:window-border (window-parent window)) c
@@ -37,11 +40,21 @@ like xterm and emacs.")
      (maximize-window window))))
 
 (defmethod window-visible-p ((window tile-window))
-  ;; In this case, visible means the window is the top window in the
-  ;; frame. This is not entirely true when it doesn't take up the
-  ;; entire frame and there's a window below it.
-  (eq window
-      (frame-window (window-frame window))))
+  ;; A TILE-WINDOW is visible is it is the top window in the frame or when the
+  ;; focused window is a FLOAT-WINDOW and the TILE-WINDOW can be seen below.
+  (let* ((frame (window-frame window))
+         (frame-windows (frame-windows (window-group window) frame)))
+    (flet ((full-frame-p (win)
+             (not (and (window-normal-hints win)
+                       (xlib:wm-size-hints-x (window-normal-hints win))
+                       (xlib:wm-size-hints-y (window-normal-hints win))))))
+      (or (eq window (or (frame-window frame)
+                         (first frame-windows)))
+          (when (> (length frame-windows) 1)
+            (loop :for (current-window next-window) :on frame-windows
+                  :until (full-frame-p current-window)
+                  :when (eq window next-window)
+                    :do (return t)))))))
 
 (defmethod window-head ((window tile-window))
   (frame-head (window-group window) (window-frame window)))
@@ -69,105 +82,116 @@ like xterm and emacs.")
 
 ;;;
 
-(defun geometry-hints (win)
-  "Return hints for max width and height and increment hints. These
+(defgeneric geometry-hints (win)
+  (:documentation "Return hints for max width and height and increment hints. These
 hints have been modified to always be defined and never be greater
-than the root window's width and height."
-  (let* ((f (window-frame win))
-         (x (frame-x f))
-         (y (frame-display-y (window-group win) f))
-         (border (xlib:drawable-border-width (window-parent win)))
-         (fwidth (- (frame-width f) (* 2 border)))
-         (fheight (- (frame-display-height (window-group win) f)
-                     (* 2 border)))
-         (width fwidth)
-         (height fheight)
-         (hints (window-normal-hints win))
-         (hints-min-width (and hints (xlib:wm-size-hints-min-width hints)))
-         (hints-min-height (and hints (xlib:wm-size-hints-min-height hints)))
-         (hints-max-width (and hints (xlib:wm-size-hints-max-width hints)))
-         (hints-max-height (and hints (xlib:wm-size-hints-max-height hints)))
-         (hints-width (and hints (xlib:wm-size-hints-base-width hints)))
-         (hints-height (and hints (xlib:wm-size-hints-base-height hints)))
-         (hints-spec-width (and hints (xlib:wm-size-hints-width hints)))
-         (hints-spec-height (and hints (xlib:wm-size-hints-height hints)))
-         (hints-inc-x (and hints (xlib:wm-size-hints-width-inc hints)))
-         (hints-inc-y (and hints (xlib:wm-size-hints-height-inc hints)))
-         (hints-min-aspect (and hints (xlib:wm-size-hints-min-aspect hints)))
-         (hints-max-aspect (and hints (xlib:wm-size-hints-max-aspect hints)))
-         (border (case *window-border-style*
-                   (:none 0)
-                   (t (default-border-width-for-type win))))
-         center)
-    ;;    (dformat 4 "hints: ~s~%" hints)
-    ;; determine what the width and height should be
-    (cond
-      ;; handle specially fullscreen windows.
-      ((window-fullscreen win)
-       (let ((head (frame-head (window-group win) f)))
-         (setf x (frame-x head)
-               y (frame-y head)
-               width (frame-width head)
-               height (frame-height head)
-               (xlib:window-priority (window-parent win)) :above))
-       (return-from geometry-hints (values x y 0 0 width height 0 t)))
-      ;; Adjust the defaults if the window is a transient_for window.
-      ((find (window-type win) '(:transient :dialog))
-       (setf center t
-             width (min (max (or hints-width 0)
-                             (or hints-min-width 0)
-                             (window-width win))
-                        width)
-             height (min (max (or hints-height 0)
-                              (or hints-min-height 0)
-                              (window-height win))
-                         height)))
-      ;; Set requested size for non-maximized windows
-      ((and (window-normal-size win)
-            hints-spec-width hints-spec-height)
-       (setf center t
-             width (min hints-spec-width width)
-             height (min hints-spec-height height)))
-      ;; aspect hints are handled similar to max size hints
-      ((and hints-min-aspect hints-max-aspect)
-       (let ((ratio (/ width height)))
-         (cond ((< ratio hints-min-aspect)
-                (setf height (ceiling width hints-min-aspect)))
-               ((> ratio hints-max-aspect)
-                (setf width  (ceiling (* height hints-max-aspect)))))
-         (setf center t)))
-      ;; Update our defaults if the window has the maxsize hints
-      ((or hints-max-width hints-max-height)
-       (when (and hints-max-width
-                  (< hints-max-width width))
-         (setf width hints-max-width))
-       (when (and hints-max-height
-                  (< hints-max-height height))
-         (setf height hints-max-height))
-       (setf center t))
-      (t
-       ;; if they have inc hints then start with the size and adjust
-       ;; based on those increments until the window fits in the frame
-       (when (and (not *ignore-wm-inc-hints*) hints-inc-x (plusp hints-inc-x))
-         (let ((w (or hints-width (window-width win))))
-           (setf width (+ w (* hints-inc-x
-                               (+ (floor (- fwidth w) hints-inc-x)))))))
-       (when (and (not *ignore-wm-inc-hints*) hints-inc-y (plusp hints-inc-y))
-         (let ((h (or hints-height (window-height win))))
-           (setf height (+ h (* hints-inc-y
-                                (+ (floor (- fheight h -1) hints-inc-y)))))))))
-    ;; adjust for gravity
-    (multiple-value-bind (wx wy) (gravity-coords (gravity-for-window win)
-                                                     width height
-                                                     0 0
-                                                     fwidth fheight)
-      (when (or center
-                (find *window-border-style* '(:tight :none)))
-        (setf x (+ wx (frame-x f))
-              y (+ wy (frame-display-y (window-group win) f))
-              wx 0 wy 0))
-      ;; Now return our findings
-      (values x y wx wy width height border center))))
+than the root window's width and height.")
+  (:method (win)
+    (let* ((f (window-frame win))
+           (x (frame-display-x (window-group win) f))
+           (y (frame-display-y (window-group win) f))
+           (border (if (or (eq *window-border-style* :none)
+                           (= (length (group-frames (window-group win))) 1))
+                       0
+                       (default-border-width-for-type win)))
+           (fwidth (- (frame-display-width (window-group win) f) (* 2 border)))
+           (fheight (- (frame-display-height (window-group win) f)
+                       (* 2 border)))
+           (width fwidth)
+           (height fheight)
+           (hints (window-normal-hints win))
+           (hints-min-width (and hints (xlib:wm-size-hints-min-width hints)))
+           (hints-min-height (and hints (xlib:wm-size-hints-min-height hints)))
+           (hints-max-width (and hints (xlib:wm-size-hints-max-width hints)))
+           (hints-max-height (and hints (xlib:wm-size-hints-max-height hints)))
+           (hints-width (and hints (xlib:wm-size-hints-base-width hints)))
+           (hints-height (and hints (xlib:wm-size-hints-base-height hints)))
+           (hints-spec-width (and hints (xlib:wm-size-hints-width hints)))
+           (hints-spec-height (and hints (xlib:wm-size-hints-height hints)))
+           (hints-inc-x (and hints (xlib:wm-size-hints-width-inc hints)))
+           (hints-inc-y (and hints (xlib:wm-size-hints-height-inc hints)))
+           (hints-min-aspect (and hints (xlib:wm-size-hints-min-aspect hints)))
+           (hints-max-aspect (and hints (xlib:wm-size-hints-max-aspect hints)))
+           center)
+      ;;    (dformat 4 "hints: ~s~%" hints)
+      ;; determine what the width and height should be
+      (cond
+        ;; handle specially fullscreen windows.
+        ((window-fullscreen win)
+         (let* ((win-group (window-group win))
+                (fs-in-frame (fullscreen-in-frame-p win))
+                (head (frame-head win-group f))
+                (frame-to-fill (if fs-in-frame f head)))
+           ;; Determine if the window should be fullscreened in the frame or the
+           ;; head. If fullscreening a frame, use the frame-display functions to
+           ;; account for the modeline and non-integer splits.
+           (if fs-in-frame
+               (setf x (frame-display-x win-group frame-to-fill)
+                     y (frame-display-y win-group frame-to-fill)
+                     width (frame-display-width win-group frame-to-fill)
+                     height (frame-display-height win-group frame-to-fill))
+               (setf x (frame-x frame-to-fill)
+                     y (frame-y frame-to-fill)
+                     width (frame-width frame-to-fill)
+                     height (frame-height frame-to-fill))))
+         (return-from geometry-hints (values x y 0 0 width height 0 t)))
+        ;; Adjust the defaults if the window is a transient_for window.
+        ((find (window-type win) '(:transient :dialog))
+         (setf center t
+               width (min (max (or hints-width 0)
+                               (or hints-min-width 0)
+                               (window-width win))
+                          width)
+               height (min (max (or hints-height 0)
+                                (or hints-min-height 0)
+                                (window-height win))
+                           height)))
+        ;; Set requested size for non-maximized windows
+        ((and (window-normal-size win)
+              hints-spec-width hints-spec-height)
+         (setf center t
+               width (min hints-spec-width width)
+               height (min hints-spec-height height)))
+        ;; aspect hints are handled similar to max size hints
+        ((and hints-min-aspect hints-max-aspect)
+         (let ((ratio (/ width height)))
+           (cond ((< ratio hints-min-aspect)
+                  (setf height (ceiling width hints-min-aspect)))
+                 ((> ratio hints-max-aspect)
+                  (setf width  (ceiling (* height hints-max-aspect)))))
+           (setf center t)))
+        ;; Update our defaults if the window has the maxsize hints
+        ((or hints-max-width hints-max-height)
+         (when (and hints-max-width
+                    (< hints-max-width width))
+           (setf width hints-max-width))
+         (when (and hints-max-height
+                    (< hints-max-height height))
+           (setf height hints-max-height))
+         (setf center t))
+        (t
+         ;; if they have inc hints then start with the size and adjust
+         ;; based on those increments until the window fits in the frame
+         (when (and (not *ignore-wm-inc-hints*) hints-inc-x (plusp hints-inc-x))
+           (let ((w (or hints-width (window-width win))))
+             (setf width (+ w (* hints-inc-x
+                                 (+ (floor (- fwidth w) hints-inc-x)))))))
+         (when (and (not *ignore-wm-inc-hints*) hints-inc-y (plusp hints-inc-y))
+           (let ((h (or hints-height (window-height win))))
+             (setf height (+ h (* hints-inc-y
+                                  (+ (floor (- fheight h) hints-inc-y)))))))))
+      ;; adjust for gravity
+      (multiple-value-bind (wx wy) (gravity-coords (gravity-for-window win)
+                                                   width height
+                                                   0 0
+                                                   fwidth fheight)
+        (when (or center
+                  (find *window-border-style* '(:tight :none)))
+          (setf x (+ wx (frame-display-x (window-group win) f))
+                y (+ wy (frame-display-y (window-group win) f))
+                wx 0 wy 0))
+        ;; Now return our findings
+        (values x y wx wy width height border center)))))
 
 (defun maximize-window (win)
   "Maximize the window."
@@ -191,43 +215,27 @@ than the root window's width and height."
           (setf (xlib:drawable-width (window-parent win)) (window-width win)
                 (xlib:drawable-height (window-parent win)) (window-height win))
           (let ((frame (window-frame win)))
-            (setf (xlib:drawable-width (window-parent win)) (- (frame-width frame)
+            (setf (xlib:drawable-width (window-parent win)) (- (frame-display-width (window-group win) frame)
                                                                (* 2 (xlib:drawable-border-width (window-parent win))))
                   (xlib:drawable-height (window-parent win)) (- (frame-display-height (window-group win) frame)
                                                                 (* 2 (xlib:drawable-border-width (window-parent win)))))))
       ;; update the "extents"
       (xlib:change-property (window-xwin win) :_NET_FRAME_EXTENTS
-                            (list wx wy
+                            (list wx
                                   (- (xlib:drawable-width (window-parent win)) width wx)
+                                  wy
                                   (- (xlib:drawable-height (window-parent win)) height wy))
                             :cardinal 32))
+    (update-decoration win)
     (update-configuration win)))
 
 ;;;
 
-(defun focus-next-window (group)
-  (focus-forward group (sort-windows group)))
+(defun only-tile-windows (windows)
+  (remove-if-not (lambda (w) (typep w 'tile-window))
+                 windows))
 
-(defun focus-prev-window (group)
-  (focus-forward group
-                 (reverse
-                  (sort-windows group))))
-
-(defcommand (next tile-group) () ()
-  "Go to the next window in the window list."
-  (let ((group (current-group)))
-    (if (group-current-window group)
-        (focus-next-window group)
-        (other-window group))))
-
-(defcommand (prev tile-group) () ()
-  "Go to the previous window in the window list."
-  (let ((group (current-group)))
-    (if (group-current-window group)
-        (focus-prev-window group)
-        (other-window group))))
-
-(defun pull-window (win &optional (to-frame (tile-group-current-frame (window-group win))))
+(defun pull-window (win &optional (to-frame (tile-group-current-frame (window-group win))) (focus-p t))
   (let ((f (window-frame win))
         (group (window-group win)))
     (unless (eq (frame-window to-frame) win)
@@ -239,7 +247,7 @@ than the root window's width and height."
       ;; We have to restore the focus after hiding.
       (when (eq win (screen-focus (window-screen win)))
         (screen-set-focus (window-screen win) win))
-      (frame-raise-window group to-frame win)
+      (frame-raise-window group to-frame win focus-p)
       ;; if win was focused in its old frame then give the old
       ;; frame the frame's last focused window.
       (when (eq (frame-window f) win)
@@ -283,7 +291,8 @@ frame."
 
 (defun other-hidden-window (group)
   "Return the last window that was accessed and that is hidden."
-  (let ((wins (remove-if (lambda (w) (eq (frame-window (window-frame w)) w)) (group-windows group))))
+  (let ((wins (remove-if (lambda (w) (eq (frame-window (window-frame w)) w))
+                         (only-tile-windows (group-windows group)))))
     (first wins)))
 
 (defun pull-other-hidden-window (group)
@@ -307,30 +316,36 @@ current frame and raise it."
 (defcommand (pull-hidden-next tile-group) () ()
 "Pull the next hidden window into the current frame."
   (let ((group (current-group)))
-    (focus-forward group (sort-windows group) t (lambda (w) (not (eq (frame-window (window-frame w)) w))))))
+    (focus-forward group (only-tile-windows (sort-windows group)) t
+                   (lambda (w) (not (eq (frame-window (window-frame w)) w))))))
 
 (defcommand (pull-hidden-previous tile-group) () ()
 "Pull the next hidden window into the current frame."
   (let ((group (current-group)))
-    (focus-forward group (nreverse (sort-windows group)) t (lambda (w) (not (eq (frame-window (window-frame w)) w))))))
+    (focus-forward group (nreverse (only-tile-windows (sort-windows group))) t
+                   (lambda (w) (not (eq (frame-window (window-frame w)) w))))))
 
 (defcommand (pull-hidden-other tile-group) () ()
 "Pull the last focused, hidden window into the current frame."
   (let ((group (current-group)))
     (pull-other-hidden-window group)))
 
-(defcommand (pull-from-windowlist tile-group) () ()
+(defcommand (pull-from-windowlist tile-group)
+    (&optional (fmt *window-format*)) (:rest)
   "Pulls a window selected from the list of windows.
 This allows a behavior similar to Emacs' switch-to-buffer
 when selecting another window."
   (let ((pulled-window (select-window-from-menu
                         (group-windows (current-group))
-                        *window-format*)))
+                        fmt)))
     (when pulled-window
       (pull-window pulled-window))))
 
-(defun exchange-windows (win1 win2)
-  "Exchange the windows in their respective frames."
+(defgeneric exchange-windows (win1 win2)
+  (:documentation "Exchange the windows in their respective frames."))
+
+(defmethod exchange-windows ((win1 tile-window) (win2 tile-window))
+  "Exchange tile windows in their respective frames."
   (let ((f1 (window-frame win1))
         (f2 (window-frame win2)))
     (unless (eq f1 f2)
@@ -363,12 +378,6 @@ when selecting another window."
 
 (defcommand-alias frame-windows echo-frame-windows)
 
-(defcommand (fullscreen tile-group) () ()
-  "Toggle the fullscreen mode of the current widnow. Use this for clients
-with broken (non-NETWM) fullscreen implemenations, such as any program
-using SDL."
-  (update-fullscreen (current-window) 2))
-
 (defcommand (gravity tile-group) (gravity) ((:gravity "Gravity: "))
   "Set a window's gravity within its frame. Gravity controls where the
 window will appear in a frame if it is smaller that the
@@ -402,9 +411,11 @@ frame. Possible values are:
   "Guess at a placement rule for WINDOW and add it to the current set."
   (let* ((group (window-group window))
          (group-name (group-name group))
-         (frame-number (frame-number (window-frame window)))
+         (frame-number-or-float (if (typep window 'float-window)
+                                    :float
+                                    (frame-number (window-frame window))))
          (role (window-role window)))
-    (push (list group-name frame-number t lock
+    (push (list group-name frame-number-or-float t lock
                 :class (window-class window)
                 :instance (window-res window)
                 :title (and title (window-name window))
@@ -424,8 +435,8 @@ frame. Possible values are:
     (if match
         (progn
           (setf *window-placement-rules* (delete match *window-placement-rules*))
-          (message "Rule forgotten"))
-        (message "No matching rule"))))
+          (message "Rule forgotten."))
+        (message "No matching rule."))))
 
 (defcommand (dump-window-placement-rules tile-group) (file) ((:rest "Filename: "))
   "Dump *window-placement-rules* to FILE."
@@ -460,11 +471,10 @@ frame. Possible values are:
                                                 (window-height-inc window)))))
       (maximize-window window)))))
 
-(defcommand (unmaximize tile-group) () ()
+(defcommand (unmaximize tile-group) (&optional (window (current-window))) (:rest)
   "Use the size the program requested for current window (if any) instead of maximizing it."
-  (let* ((window (current-window))
-         (status (not (window-normal-size window)))
-         (hints (window-normal-hints window)))
+  (let ((status (not (window-normal-size window)))
+        (hints (window-normal-hints window)))
     (if (and (xlib:wm-size-hints-width hints)
              (xlib:wm-size-hints-height hints))
         (progn
@@ -478,10 +488,10 @@ frame. Possible values are:
 frame and focus the selected window.  The optional argument @var{fmt} can be
 specified to override the default window formatting."
   (let* ((group (current-group))
-	 (frame (tile-group-current-frame group)))
+         (frame (tile-group-current-frame group)))
     (if (null (frame-windows group frame))
-	(message "No Managed Windows")
-	(let ((window (select-window-from-menu (frame-sort-windows group frame) fmt)))
-	  (if window
-	      (group-focus-window group window)
-	      (throw 'error :abort))))))
+        (message "No Managed Windows.")
+        (let ((window (select-window-from-menu (frame-sort-windows group frame) fmt)))
+          (if window
+              (group-focus-window group window)
+              (throw 'error :abort))))))
